"""

    PeakMapper.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF

"""
import copy
from bisect import bisect_right
import numpy            as np
import logging
from scipy.stats        import pearsonr, linregress
from scipy.interpolate  import UnivariateSpline
from scipy.optimize     import minimize
from itertools          import combinations
import molass_legacy.KekLib.OurStatsModels as sm
from importlib import reload
from .SimplestMapper import SimplestMapper
from .PeakCurve import PeakCurve, get_peak_curve_info
from molass_legacy.SerialAnalyzer.ElutionCurve       import ElutionCurve
from molass_legacy.Elution.CurveUtils  import simple_plot, find_rotated_extreme_arg
from molass_legacy.KekLib.SciPyCookbook      import smooth
from molass_legacy.KekLib.BasicUtils         import Struct
from molass_legacy.KekLib.ExceptionTracebacker import log_exception, warnlog_exception
import molass_legacy.KekLib.DebugPlot        as plt
from molass_legacy._MOLASS.SerialSettings     import get_setting

COMPARE_ADDING_PEAKS_2WAYS  = True
NUM_PEAKS_SCORING_WEIGHT    = 0.01
USE_SCORED_BOUNDARY         = False
ROUGH_MIN_Y_RATIO_LIST      = [0.02, 0.1]
ADEQUACY_CORRELATION        = 0.997     # 0.996 for monoE, 0.99997 for SUB_TRN1
RISKY_CORRELATION_LIMIT     = 0.95      # > 0.88 for 20190221_2
NUM_ROOTS_TO_AVOID_NOISE    = 5         # to void low quality peaks as in pH7, > 3 for SUB_TRN1
FEASIBLE_DIFF_NUM           = 2
REDUCE_WHEN_INFEASIBLE      = True
ADOPTABLE_SCORE_RATIO       = 1.05      # > 1.01 for 20170304
RELIABLE_PEAK_SCORE         = 0.9
NUM_FRAMES_BOUNADRY         = 400       # < 461 for 20210628_2
MINIMUM_HEIGHT_RATIO        = 0.04      # > 0.03 for 20200121_2
ACCEPTABLE_SIMILARITY       = 0.97      # > 0.93 for 20210323_1, 0.96 for 20200121_2
DEEM_SINGLE_PEAK_RATIO      = 0.1
ACCEPTABLE_ENDS_RATIO_DEV   = 0.5

def debug_plot( a_curve, x_curve, where_str, params=None):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

    plt.push()
    fig = plt.figure(figsize=(14, 7))
    fig.suptitle( 'PeakMapper debug plot for ' + get_in_folder() + ' (%s)' % where_str, fontsize=24 )
    ax1 = fig.add_subplot( 121 )
    ax2 = fig.add_subplot( 122 )
    ax1.set_title( "UV Curve", fontsize=20 )
    ax2.set_title( "Xray Curve", fontsize=20 )

    ax1.plot( a_curve.y )
    ax1.plot( a_curve.spline(a_curve.x), color='lime' )
    ax2.plot( x_curve.y, color='orange' )
    ax2.plot( x_curve.spline(x_curve.x), color='lime' )

    def plot_peaks( ax, curve ):
        for info in curve.peak_info:
            peak = info[1]
            ax.plot( peak, curve.spline( peak ), 'o', color='red', markersize=10 )

        for b in curve.boundaries:
            ax.plot( b, curve.spline( b ), 'o', color='yellow', markersize=8  )

    plot_peaks( ax1, a_curve )
    plot_peaks( ax2, x_curve )

    if params is not None:
        A, B = params
        A_, B_ = 1/A, -B/A
        for x1 in a_curve.peak_top_x:
            x = A_*x1 + B_
            y = x_curve.spline(x)
            ax2.plot(x, y, 'o', color='cyan')

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()
    plt.pop()

def simple_plot_pair_curves(title, a_curve, x_curve):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
    in_folder = get_in_folder()

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle(title + " with %s" % in_folder, fontsize=20)
        simple_plot(ax1, a_curve, color="blue")
        simple_plot(ax2, x_curve, color="orange")
        fig.tight_layout()
        plt.show()

def mapping_proof(title, a_curve, x_curve, mapping):
    x = x_curve.x
    y = x_curve.y
    A, B = mapping
    uv_x = x*A + B

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle(title)

        simple_plot(ax1, a_curve, legend=False)
        simple_plot(ax2, x_curve)

        ax1.plot(uv_x, y, label="mapped")
        ax1.legend()

        fig.tight_layout()
        plt.show()

def make_unique_peak_info( old_list ):
    temp_list = sorted( old_list )

    new_list = []
    last_info = None
    for info in temp_list:
        if last_info is None or info[0] != last_info[0]:
            new_list.append( info )

        last_info = info

    return new_list

def make_unique_boundaries( old_list ):
    return sorted( list( set( old_list ) ) )

class ProxyCurve(ElutionCurve):
    def __init__( self, curve, peak_top_x, peak_info, boundaries, add_peak_info, add_boundaries, correct=False ):
        self.__dict__   = curve.__dict__.copy()
        self.logger     = logging.getLogger(__name__)
        self.peak_top_x = np.array(peak_top_x)
        self.peak_info  = peak_info
        self.boundaries = boundaries
        self.peak_slice = slice(0, len(curve.x))    # comptibility with PeakCurve
        self.emg_peaks = None
        if correct:
            self.correct_valley_bottm(debug=True)

    def correct_valley_bottm(self, debug=False):

        if debug:
            saved_boundaries = copy.deepcopy(self.boundaries)

        for k, b in enumerate(self.boundaries):
            start = max(0, b - 20)
            stop  = min(len(self.x), b + 20)
            y_ = self.spline(self.x[start:stop])
            ppp = np.argpartition(y_, 5)
            n = start + int(np.average(ppp[0:5+1]) + 0.5)
            if n != b:
                self.logger.info("corrected %dth valley bottm from %d to %d." % (k, b, n))

        if debug:
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(saved_boundaries, self.spline(saved_boundaries), 'o', color='gray', alpha=0.2, markersize=10, label='saved_boundaries')
            simple_plot(ax, self, "correct_valley_bottm debug")
            fig.tight_layout()
            plt.show()

    def get_peak_region_width(self):
        # note that ElutionCurve does not currently have this method
        return self.peak_slice.stop - self.peak_slice.start

class MajorCurve(ElutionCurve):
    def __init__(self, curve):
        self.__dict__   = curve.__dict__.copy()
        self.logger     = logging.getLogger(__name__)
        self.peak_top_x = curve.major_peak_top_x
        self.peak_info  = curve.major_peak_info
        self.boundaries = curve.major_boundaries

class LessPeakCurve(ElutionCurve):
    def __init__(self, curve, num_less):
        self.__dict__   = curve.__dict__.copy()
        self.logger     = logging.getLogger(__name__)
        peak_height_ratios = [ curve.get_height_ratio(top) for top in curve.peak_top_x ]
        less_index = np.argpartition(peak_height_ratios, num_less)[:num_less]
        print('peak_height_ratios=', peak_height_ratios)
        print('less_index=', less_index)
        new_peak_top_x = []
        for k, top in enumerate(curve.peak_top_x):
            if k in less_index:
                continue
            new_peak_top_x.append(top)
        self.peak_top_x = np.array(new_peak_top_x)
        self.set_peak_dependent_info()

class MappedInfo(list):
    def __init__(self, info_list, curves=None):
        list.__init__(self, info_list)
        self._curves = curves
        best_params = info_list[0]
        if best_params is None:
            # this can happen in the preliminary recongnition as in 20171226
            pass
        else:
            self._A, self._B = best_params

    def xr_to_uv(self, xr_p):
        return self._A * xr_p + self._B

    def uv_to_xr(self, uv_p):
        return (uv_p - self._B)/self._A

def slice_length(s):
    return s.stop - s.start

class PeakMapper:
    def __init__(self, a_curve, x_curve, simply=False, pre_recog=None, debug=False, debug_detail=False ):
        self.logger = logging.getLogger(__name__)

        self.a_curve = a_curve
        self.x_curve = x_curve
        self.peak_mapping_only = get_setting('peak_mapping_only')
        self.feature_mapped = False
        self.reliable = True
        self.debug  = debug
        self.debug_detail = debug_detail

        fallback = False
        if not simply:
            try:
                sliced_pm_info = self.get_sliced_peak_mapper(a_curve, x_curve, debug=debug)
                self.pre_recog  = pre_recog
                self.make_it_compatible_from_sliced(sliced_pm_info, debug=debug)
            except Exception as exc:
                warnlog_exception(self.logger, "sliced peak mapping failed: ")
                simply = True
                fallback = True
                self.logger.warning("resorting to the simple treatment (i.e., without slicing)  due to the previous error.")

        if simply:
            if debug:
                simple_plot_pair_curves("PeakCurves (0)", self.a_curve, self.x_curve)

            try:
                self.pre_recog = None
                self.do_mapping(debug=debug)
                if fallback:
                    A, B =self.get_imroved_mapping(fallback=fallback)
                    self.mapped_info[0] = (A, B)
                    self.best_info[1] = (A, B)
            except Exception as exc:
                # as in 20201007_2
                from .ReducedCurve import make_reduced_curve
                warnlog_exception(self.logger, "get_imroved_mapping failed: ")
                self.logger.warning("resorting to curve reducing due to exception: %s", str(exc))
                curves = []
                for curve in self.a_curve, self.x_curve:
                    curves.append(make_reduced_curve(curve))
                self.do_mapping(reduced_curves=curves, debug=debug)

    def preliminary_check(self, a_curve, x_curve):
        a_frames = len(a_curve.x)
        x_frames = len(x_curve.x)
        frames_ratio = a_frames/x_frames
        almost_same_scale = abs(frames_ratio - 1) < 0.1

        peaktop_ratio = (a_curve.primary_peak_x/a_frames)/(x_curve.primary_peak_x/x_frames)
        proportinal_peaktop = abs(peaktop_ratio - 1) < 0.1

        a_peaks = len(a_curve.peak_info)
        x_peaks = len(x_curve.peak_info)
        same_few_num_peaks = a_peaks == x_peaks and x_peaks <= 2

        return Struct(  almost_same_scale=almost_same_scale,
                        frames_ratio=frames_ratio,
                        proportinal_peaktop=proportinal_peaktop,
                        a_peaks = a_peaks,
                        x_peaks = x_peaks,
                        same_few_num_peaks=same_few_num_peaks,
                        )

    def check_peak_existence(self, a_curve, x_curve, title, plot=True):
        for curve in a_curve, x_curve:
            if len(curve.peak_info) == 0:
                if plot:
                    with plt.Dp():
                        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                        fig.suptitle(title)
                        simple_plot(ax1, a_curve, color="blue")
                        simple_plot(ax2, x_curve, color="orange")
                        fig.tight_layout()
                        plt.show()
                assert False

    def get_sliced_peak_mapper(self, a_curve, x_curve, debug=False):
        self.check_peak_existence(a_curve, x_curve, "get_sliced_peak_mapper entry")

        check_info = self.preliminary_check(a_curve, x_curve)
        if debug:
            print("check_info=", check_info)
            simple_plot_pair_curves("get_sliced_peak_mapper (1)", a_curve, x_curve)

        peak_curve_ok = False
        if len(x_curve.x) >= NUM_FRAMES_BOUNADRY:
            try:
                a_curve_, a_slice_ = get_peak_curve_info(a_curve, debug=debug)
                x_curve_, x_slice_ = get_peak_curve_info(x_curve, debug=debug)
                try:
                    a_curve_, a_slice_, x_curve_, x_slice_ = self.improve_slices_if_possible(check_info, a_curve_, a_slice_, x_curve_, x_slice_)
                except:
                    warnlog_exception(self.logger, "improve_slices_if_possible failed: ")
                peak_curve_ok = True
            except:
                # as in 20201127_5
                warnlog_exception(self.logger, "get_sliced_peak_mapper(1)")
        elif len(a_curve.x) >= NUM_FRAMES_BOUNADRY*2:
            # as in 20201009_1
            try:
                a_curve_, a_slice_ = get_peak_curve_info(a_curve, debug=debug)
                x_curve_, x_slice_ = x_curve, None
                peak_curve_ok = True
            except:
                warnlog_exception(self.logger, "get_sliced_peak_mapper (2)")

        if not peak_curve_ok:
            # as in 20191114_3 where len(a_curve.x) >= 1000 while len(x_curve.x) < 1000
            # avoiding the inconvenience in get_peak_curve_info, which should better be investigated anyway
            a_curve_, a_slice_ = a_curve, None
            x_curve_, x_slice_ = x_curve, None

        if debug:
            simple_plot_pair_curves("get_sliced_peak_mapper (2)", a_curve_, x_curve_)

        self.check_peak_existence(a_curve_, x_curve_, "get_sliced_peak_mapper before recursive construction")

        pm = PeakMapper(a_curve_, x_curve_, simply=True, debug=debug)
        return Struct(pm=pm, a_slice_=a_slice_, x_slice_=x_slice_)

    def improve_slices_if_possible(self, check_info, a_curve_, a_slice_, x_curve_, x_slice_, debug=False):
        sliced_check_info = self.preliminary_check(a_curve_, x_curve_)
        # print("------------------- check_info=", check_info)
        # print("------------------- sliced_check_info=", sliced_check_info)
        if sliced_check_info.a_peaks > 2 and sliced_check_info.x_peaks >= 2:
            # as in 20220716/FER_OA_302
            return a_curve_, a_slice_, x_curve_, x_slice_

        if check_info.almost_same_scale:
            pass
        else:
            if check_info.proportinal_peaktop:
                pass
            else:
                ok = True
                if check_info.same_few_num_peaks:
                    if not sliced_check_info.same_few_num_peaks:
                        # as in 20200121_3
                        ok = False
                if ok:
                    return a_curve_, a_slice_, x_curve_, x_slice_
                else:
                    pass

        ratio = slice_length(a_slice_)/(check_info.frames_ratio * slice_length(x_slice_))
        if abs(ratio - 1) < 0.1:
            return a_curve_, a_slice_, x_curve_, x_slice_

        """
        try to improve the slice in cases like 20200121_2 or 20200630_11
        """

        if not check_info.same_few_num_peaks:
            # like 20200630_11
            # note that unmatching peaks will be removed
            num_ratio = len(a_curve_.peak_info)/len(x_curve_.peak_info)
            # assert (ratio - 1)*(num_ratio - 1) > 0      # i.e. of the same sign
            curve_name = "x_curve" if num_ratio < 1 else "a_curve"
            self.logger.info("unmatching peaks will probably be removed from %s", curve_name)

        if ratio < 1:
            curve1 = self.a_curve
            curve1_ = a_curve_
            slice1_ = a_slice_
            curve2 = self.x_curve
            curve2_ = x_curve_
            slice2_ = x_slice_
            self.logger.info("improve_slices_if_possible: improving x_slice_=%s from a_slice_=%s for ratio=%.3g", str(x_slice_), str(a_slice_), ratio)
        else:
            curve1 = self.x_curve
            curve1_ = x_curve_
            slice1_ = x_slice_
            curve2 = self.a_curve
            curve2_ = a_curve_
            slice2_ = a_slice_
            self.logger.info("improve_slices_if_possible: improving a_slice_=%s from x_slice_=%s for ratio=%.3g", str(a_slice_), str(x_slice_), ratio)

        sm = SimplestMapper(curve1, curve2)
        slice_ = slice(sm.map_1to2(slice1_.start), sm.map_1to2(slice1_.stop))
        ecurve = ElutionCurve(curve2.y[slice_])
        ecurve.get_default_editor_ranges()

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("improve_slices_if_possible result")
                simple_plot(ax1, curve1)
                simple_plot(ax2, curve2, legend=False)
                ax2.plot(slice_.start+ecurve.x, ecurve.y, ":", label="with improved slice")
                ax2.legend()
                fig.tight_layout()
                plt.show()

        if ratio < 1:
            return curve1_, slice1_, ecurve, slice_
        else:
            return ecurve, slice_, curve1_, slice1_

    def check_minimum_mapping_quality(self, a_curve, x_curve, where_text):
        x = [0]
        y = [0]
        for info1, info2 in zip(a_curve.peak_info, x_curve.peak_info):
            x.append(info2[1])
            y.append(info1[1])
        x.append(len(x_curve.x))
        y.append(len(a_curve.x))

        correlation = pearsonr(x, y)[0]
        print(where_text, "------------------ correlation=", correlation)

    def do_mapping(self, reduced_curves=None, debug=False):

        if reduced_curves is None:
            a_curve = self.a_curve
            x_curve = self.x_curve
        else:
            a_curve, x_curve = reduced_curves

        if debug:
            simple_plot_pair_curves("before mapping", a_curve, x_curve)

        if self.debug:
            from molass_legacy.KekLib.CallStack import CallStack
            cstack = CallStack()
            print('PeakMapper call stack=', cstack)
            debug_plot( a_curve, x_curve, "before" )

        A, B = self.get_simplest_params(a_curve, x_curve)

        if debug:
            import Mapping.PeakMappingSolver
            from importlib import reload
            reload(Mapping.PeakMappingSolver)
        from .PeakMappingSolver import PeakMappingSolver
        pms = PeakMappingSolver(a_curve, x_curve, (A, B), debug=debug)
        uv_mapper, xr_mapper = pms.get_opt_mappers()
        try:
            better_a_curve = self.make_better_curve(a_curve, x_curve, xr_mapper, (A, B), "better_a_curve", debug=debug)
            better_x_curve = self.make_better_curve(x_curve, a_curve, uv_mapper, (1/A, -B/A), "better_x_curve", debug=debug)
            curves = [better_a_curve, better_x_curve]
        except Exception as exc:
            warnlog_exception(self.logger, "making thinner curves due to ")
            thinner_a_curve = self.make_thinner_curve(a_curve, uv_mapper)
            thinner_x_curve = self.make_thinner_curve(x_curve, xr_mapper)
            curves = [thinner_a_curve, thinner_x_curve]

        if self.debug:
            debug_plot(*curves, "after")

        self.check_peak_existence(*curves, "do_mapping result", plot=False)
        self.check_minimum_mapping_quality(*curves, "do_mapping result")

        self.mapped_curves = curves

        if debug:
            simple_plot_pair_curves("after mapping", *curves)

    def make_it_compatible_from_sliced(self, sliced_pm_info, debug=False):
        pm = sliced_pm_info.pm
        a_slice_ = sliced_pm_info.a_slice_
        a_start = 0 if a_slice_ is None else a_slice_.start
        x_slice_ = sliced_pm_info.x_slice_
        x_start = 0 if x_slice_ is None else x_slice_.start
        mapped_curves = pm.get_mapped_curves()
        # A, B = pm.best_info[1]
        A, B = pm.get_imroved_mapping()

        B_ = a_start + A*(-x_start) + B

        if debug:
            simple_plot_pair_curves("make_it_compatible_from_sliced", *mapped_curves)

        # None's will not be used
        self.best_info = [pm.best_info[0], (A,B_), None, None, None]
        self.mapped_info = MappedInfo([(A, B_), [], None], [self.a_curve, self.x_curve])
        curves = []
        for curve, mp_curve, slice_ in [(self.a_curve, mapped_curves[0], a_slice_),
                                        (self.x_curve, mapped_curves[1], x_slice_) ]:
            if slice_ is None:
                pkcurve = mp_curve
            else:
                pkcurve = PeakCurve(curve, pkslice_info=(mp_curve, slice_))
                pkcurve.get_emg_peaks()

                """
                try:
                    pkcurve = PeakCurve(curve, pkslice_info=(mp_curve, slice_))
                    pkcurve.get_emg_peaks()
                except:
                    warnlog_exception(self.logger, "make_it_compatible_from_sliced: ")
                    pkcurve = curve
                """

            curves.append(pkcurve)

        self.check_minimum_mapping_quality(*curves, "make_it_compatible_from_sliced result")
        self.mapped_curves = curves

        if debug:
            simple_plot_pair_curves("after make_it_compatible_from_sliced", *curves)

    def get_simplest_params(self, curve1, curve2):
        max_score, max_params, max_indeces, max_correl, max_simil, max_info = self.get_best_mapping_params(curve1, curve2, peak_mapping_only=False)
        B, A = max_params

        if len(curve1.peak_info) < 3 or len(curve2.peak_info) < 3:
            a, b = self.compare_primary_peak_mapping_params(curve1, curve2, (A, B))
        else:
            # avoid wornsening as with 202300705 (BSA)
            a, b = A, B

        # for backward compatibility begin
        self.best_info = [max_score, (a, b), max_indeces, max_correl, max_simil]
        self.mapped_info = MappedInfo([(a, b), [max_correl, max_simil], max_info], [self.a_curve, self.x_curve])
        # for backward compatibility end

        return a, b

    def get_mapped_curves( self ):
        return self.mapped_curves

    def find_best_mapping( self, curve1, curve2, check_adequacy=True, features=True):

        curve1_p, curve2_p, best_info_p, mapped_info_p = self.find_best_mapping_impl(curve1, curve2, False, check_adequacy, debug=False)

        if features and not self.peak_mapping_only:
            curve1_f, curve2_f, best_info_f, mapped_info_f = self.find_best_mapping_impl(curve1_p, curve2_p, True, check_adequacy, mapped_info=mapped_info_p, debug=False)

            if best_info_f[0] > best_info_p[0]:
                best_info = best_info_f
                mapped_info = mapped_info_f
                ret_curves = [curve1_f, curve2_f]
                self.feature_mapped = True
                self.logger.info("features matching has been successful: %g > %g" % (best_info_f[0],  best_info_p[0]))
            else:
                best_info = best_info_p
                mapped_info = mapped_info_p
                ret_curves = [curve1_p, curve2_p]

                if self.pre_recog is not None:
                    mapped_info = self.pre_recog.get_restricted_mapped_info()
                    curve1_f, curve2_f, best_info_f, mapped_info_f = self.find_best_mapping_impl(curve1_p, curve2_p, True, check_adequacy, mapped_info=mapped_info, debug=False)
                    if best_info_f[0] > best_info_p[0]:
                        best_info = best_info_f
                        mapped_info = mapped_info_f
                        ret_curves = [curve1_f, curve2_f]
                        self.feature_mapped = True

                if self.feature_mapped:
                    self.logger.info("features matching has been successful using pre-recog: %g > %g" % (best_info_f[0],  best_info_p[0]))
                else:
                    self.logger.warning("features matching has been worse: %g <= %g" % (best_info_f[0], best_info_p[0]))
        else:
            best_info = best_info_p
            mapped_info = mapped_info_p
            ret_curves = [curve1_p, curve2_p]

        self.mapped_info = mapped_info
        self.best_info = best_info

        return ret_curves

    def evaluate_mapping_impl( self, size1, size2, curve1, peaks1, index1, curve2, peaks2, index2, use_ends=True, debug=False):
        if use_ends and (len(peaks1) < 2 or len(peaks2) < 2):
            y   = [ 0, size1-1 ]
            x   = [ 0, size2-1 ]
            w   = [ 1, 1 ]
        else:
            y   = []
            x   = []
            w   = []

        for p1, p2 in zip( peaks1[index1], peaks2[index2] ):
            y.append( p1 )
            x.append( p2 )
            w.append( 10 )

        if not use_ends:
            for rec1, rec2 in zip(curve1.peak_info, curve2.peak_info):
                for i in [0, 2]:
                    y.append( rec1[i] )
                    x.append( rec2[i] )
                    w.append( 5 )

        X   = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        res = mod.fit()

        if debug:
            B, A = res.params
            mp_x = curve2.x*A + B
            mp_y = curve2.spline(mp_x)/curve2.max_y*curve1.max_y
            rmsd = np.sqrt(np.average((curve1.spline(mp_x) - mp_y)**2))

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("evaluate_mapping_impl")
                simple_plot(ax1, curve1, color="blue")
                ax1.plot(mp_x, mp_y, ":")
                simple_plot(ax2, curve2, color="orange")
                fig.tight_layout()
                plt.show()

        correl, simil = self.compute_mapped_curve_simimarity(res.params, x, y, w, [index1, index2], curve1, curve2, debug=False )

        return correl, simil, res.params, [x, y, w]

    def evaluate_mapping( self, size1, size2, curve1, peaks1, index1, curve2, peaks2, index2):

        first_result = self.evaluate_mapping_impl(size1, size2, curve1, peaks1, index1, curve2, peaks2, index2)
        correl, simil, params, info = first_result
        first_score = correl + simil
        # print('first_score=', first_score, (correl, simil) )
        if len(curve1.peak_info) == 1 and len(curve2.peak_info) == 1 and simil < 0.9:
            try:
                second_result = self.evaluate_mapping_impl(size1, size2, curve1, peaks1, index1, curve2, peaks2, index2, use_ends=False)
                correl, simil, params, info = second_result
                second_score = correl + simil
                # print('second_score=', second_score, (correl, simil) )
            except:
                second_score = 0
            score_ratio = second_score/first_score
            print('score_ratio=', score_ratio)
            if score_ratio > ADOPTABLE_SCORE_RATIO:
                # there should be enough superiority to adopt the non-use_ends result
                self.logger.info("the 2nd result scored %.3g has been adopted over the 1st result scored %.3g." % (second_score, first_score) )
                ret_result = second_result
            else:
                ret_result = first_result
        else:
            ret_result = first_result

        return ret_result

    def evaluate_mapping_features( self, curve1, points1, weights1, index1, curve2, points2, weights2, index2):
        y = []
        x = []
        w = []

        assert_later = False

        inappropriate = False
        for p1, w1, p2, w2 in zip( points1[index1], weights1[index1], points2[index2], weights2[index2] ):
            y.append( p1 )
            x.append( p2 )
            if assert_later:
                if abs(w1 - w2) > 1e-6:
                    inappropriate = True
            else:
                # debug_plot(curve1, curve2, "evaluate_mapping_features")
                assert abs(w1 - w2) < 1e-6
            w.append( max(w1, w2) )

        X   = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        res = mod.fit()

        correl, simil = self.compute_mapped_curve_simimarity(res.params, x, y, w, [index1, index2], curve1, curve2, features=True, debug=False )

        if assert_later:
            if inappropriate:
                assert False
                # pass

        return correl, simil, res.params, [x, y, w]

    def compute_mapped_curve_simimarity(self, params, x, y, w, indeces, curve1, curve2, features=False, nodebug=False, debug=False):
        intercept, slope = params

        correlation = pearsonr(x, y)[0]

        j_ = np.array([0, len(curve1.x)])
        i_ = 1/slope*j_ - intercept/slope
        i_start = int( min(0, i_[0]) )
        i_stop = int( max(len(curve2.x), i_[-1]))

        i = np.arange(i_start, i_stop)
        j = slope * i + intercept

        source_y = curve2.spline(i)
        try:
            scale = curve2.height/curve1.height
        except:
            self.logger.warning("curve height not available, using max_y instead.")
            scale = curve2.max_y/curve1.max_y
        mapped_y_ = curve1.spline(j) * scale

        def obj_func(params):
            d = np.sum(np.abs(mapped_y_ + params[0] - source_y))
            return d

        # allow any vertical shift and minimize the gap area
        result = minimize(obj_func, np.zeros(1))

        diff_area = obj_func(result.x)
        mapped_y = mapped_y_ + result.x[0]

        whole_area = curve2.height * len(source_y)
        similarity = abs(1 -  diff_area/whole_area)

        if debug:
        # if True:
            from matplotlib.patches import Polygon
            from  molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

            in_folder = get_in_folder()
            markersize = 10

            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21,7))
            ax1, ax2, ax3 = axes
            ax1t = ax1.twinx()

            fig.suptitle("Mapping Explained for %s: indeces=%s" % (in_folder, str(indeces)), fontsize=20)
            ax1.set_title("Data Curves", fontsize=16)
            ax2.set_title("Elution No Mapping", fontsize=16)
            ax3.set_title("Mapped Curves", fontsize=16) 

            c_colors = ['blue', 'orange']
            p_colors = ['cyan', 'orange']

            ax1.plot(j, curve1.spline(j), color=c_colors[0])
            ax1t.plot(i, source_y, color=c_colors[1])

            ax2.plot(i, j, color='green')
            xmin2, xmax2 = ax2.get_xlim()
            ymin2, ymax2 = ax2.get_ylim()

            spanx = xmax2 - xmin2
            spany = ymax2 - ymin2
            xmin_ = -spanx*0.1
            ymin_ = -spany*0.1

            ax2.plot(x, [ymin_]*len(x), 'o', markersize=markersize, color=p_colors[1])
            ax2.plot([xmin_]*len(y), y, 'o', markersize=markersize, color=p_colors[0])

            for px, py, pw in zip(x, y, w):
                ax1.plot(py, curve1.spline(py), 'o', markersize=markersize, color=p_colors[0])
                ax1t.plot(px, curve2.spline(px), 'o', markersize=markersize, color=p_colors[1])
                ax2.plot(px, py, 'o', markersize=markersize, color='lime')
                py_ = slope * px + intercept
                ax3.plot(px, curve1.spline(py_) * scale, 'o', markersize=markersize, color=p_colors[0])
                ax3.plot(px, curve2.spline(px), 'o', markersize=markersize, color=p_colors[1])
                shrinky = spany*0.02/(py - ymin_)
                ax2.annotate("", xy=(px, py), xytext=(px, ymin_), alpha=0.5,
                            arrowprops=dict(arrowstyle='->', ls='dashed', lw=1.5, color=c_colors[1], shrinkA=10, shrinkB=10),
                            )
                shrinkx = spanx*0.02/(px - xmin_)
                ax2.annotate("", xy=(xmin_, py), xytext=(px, py), alpha=0.5,
                            arrowprops=dict(arrowstyle='->', ls='dashed', lw=1.5, color=c_colors[0], shrinkA=10, shrinkB=10),
                            )

                if pw > 0.9:
                    ax1.plot(py, curve1.spline(py), 'o', markersize=5, color='red')
                    ax1t.plot(px, curve2.spline(px), 'o', markersize=5, color='red')
                    ax3.plot(px, curve1.spline(py_) * scale, 'o', markersize=5, color='red')
                    ax3.plot(px, curve2.spline(px), 'o', markersize=5, color='red')

            ax3.plot(i, source_y, color=c_colors[1])
            ax3.plot(i, mapped_y, color=c_colors[0])
            ax3.plot(i, mapped_y_, ':', color=c_colors[0])

            poly_points = list(zip( i, source_y )) + list( reversed( list( zip( i, mapped_y ) ) ) )
            diff_poly   = Polygon( poly_points, alpha=0.1 )
            ax3.add_patch(diff_poly)

            xmin, xmax = ax2.get_xlim()
            ymin, ymax = ax2.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax2.text(tx, ty, "Linearity=%.3g" % correlation, ha='center', va='center', fontsize=50, alpha=0.1)

            xmin, xmax = ax3.get_xlim()
            ymin, ymax = ax3.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax3.text(tx, ty, "Similarity=%.3g" % similarity, ha='center', va='center', fontsize=50, alpha=0.1)

            fig.tight_layout()
            fig.subplots_adjust(top=0.85)
            plt.show()
            plt.pop()

        if False and not nodebug and self.debug_detail or debug:
            from scipy.interpolate import UnivariateSpline
            from matplotlib.patches import Polygon

            colors = ['blue', 'orange']

            size1   = len(curve1.y)
            size2   = len(curve2.y)

            if features:
                points1, weights1 = curve1.feature.get_points()
                points2, weights2 = curve2.feature.get_points()
                peaks1  = points1
                peaks2  = points2
                points_label = 'points'
                weight_boundary = 0.75
            else:
                peaks1  = np.array( [ info[1] for info in curve1.peak_info ] )
                peaks2  = np.array( [ info[1] for info in curve2.peak_info ] )
                points_label = 'peaks'
                weight_boundary = 5

            unmapped_index = []
            for k in range(len(peaks1)):
                if k not in indeces:
                    unmapped_index.append(k)

            mapped_spline = UnivariateSpline(i, mapped_y, s=0, ext=3)
            mapped_peaks = 1/slope*peaks1 - intercept/slope

            fig = plt.figure(figsize=(14,7))
            fig.suptitle("Debug: ac_vector: indeces=" + str(indeces))

            ax1 = fig.add_subplot(121)
            ax2 = fig.add_subplot(122)
            ax1.set_title("Index Mapping") 
            ax2.set_title("Mapped Curves") 

            x_  = [ 0, size2-1 ]
            y_  = [ slope*z + intercept  for z in x_ ]
            ax1.plot( x_, y_, color='green' )

            for px, py, pw in zip(x, y, w):
                # color = 'red' if pw > weight_boundary else 'black'
                color = 'red'
                msize = 10 if pw > weight_boundary else 5
                ax1.plot(px, py, 'o', color=color, markersize=msize)

            ax2.plot(i, source_y, color=colors[1], label='source')
            ax2.plot(i, mapped_y, color=colors[0], label='mapped')
            ax2.plot(i, mapped_y_, ':', color=colors[0], label='mapped')

            if len(unmapped_index) > 0:
                mapped_peaks_ = mapped_peaks[unmapped_index]
                ax2.plot(mapped_peaks_, mapped_spline(mapped_peaks_), 'o', color='yellow', markersize=10, label='unmapped ' + points_label)

            """
            TO BE FIXED:
            there was an error occured in 20170511
            IndexError: index 1 is out of bounds for axis 0 with size 1
            """
            print('mapped_peaks=', mapped_peaks, 'indeces=', indeces)
            mapped_peaks_ = mapped_peaks[indeces[0]]
            ax2.plot(mapped_peaks_, mapped_spline(mapped_peaks_), 'o', color='red', markersize=10, label='mapped ' + points_label)

            ax2.plot(peaks2, curve2.spline(peaks2), 'o', color='pink', markersize=10, label='source ' + points_label)

            poly_points = list(zip( i, source_y )) + list( reversed( list( zip( i, mapped_y ) ) ) )
            diff_poly   = Polygon( poly_points, alpha=0.2 )
            ax2.add_patch(diff_poly)

            xmin, xmax = ax1.get_xlim()
            ymin, ymax = ax1.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax1.text(tx, ty, "Linearity=%.3g" % correlation, ha='center', va='center', fontsize=60, alpha=0.2)

            xmin, xmax = ax2.get_xlim()
            ymin, ymax = ax2.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax2.text(tx, ty, "Similarity=%.3g" % similarity, ha='center', va='center', fontsize=60, alpha=0.2)

            ax2.legend()
            fig.tight_layout()
            fig.subplots_adjust(top=0.9)
            plt.show()

        return correlation, similarity

    def get_best_mapping_params(self, curve1, curve2, peak_mapping_only=False):
        if peak_mapping_only:
            peaks1  = np.array( [ info[1] for info in curve1.get_major_peak_info(ratio=0.3) ] )     # ratio=0.1 is too small for 20190221_2
            peaks2  = np.array( [ info[1] for info in curve2.get_major_peak_info(ratio=0.3) ] )     # 
        else:
            peaks1  = np.array( [ info[1] for info in curve1.peak_info ] )
            peaks2  = np.array( [ info[1] for info in curve2.peak_info ] )

        size1   = len(curve1.y)
        size2   = len(curve2.y)
        num_peaks1 = len(peaks1)
        num_peaks2 = len(peaks2)

        greater_num = max(num_peaks1, num_peaks2)
        less_num = min(num_peaks1, num_peaks2)

        greater_index = list(range(greater_num))
        full_index1 = list(range(num_peaks1))
        full_index2 = list(range(num_peaks2))

        max_score   = None
        max_params  = None
        max_indeces = None
        max_correl  = None
        max_simil   = None
        max_info    = None

        for c in combinations(greater_index, less_num):
            index = list(c)
            if num_peaks1 >= num_peaks2:
                index1 = index
                index2 = full_index2
            else:
                index1 = full_index1
                index2 = index
            assert len(index1) == len(index2)
            correl, simil, params, info = self.evaluate_mapping( size1, size2, curve1, peaks1, index1, curve2, peaks2, index2 )
            score = correl + simil
            if max_score is None or score > max_score:
                max_score   = score
                max_params  = params
                max_indeces = (index1, index2)
                max_correl  = correl
                max_simil   = simil
                max_info    = info

        return max_score, max_params, max_indeces, max_correl, max_simil, max_info

    def get_best_mapping_features_params( self, curve1, curve2, mapped_info ):

        major_peak_info1 = curve1.get_major_peak_info()
        major_peak_info2 = curve2.get_major_peak_info()

        adopted_peak_info1 = []
        adopted_peak_info2 = []
        best_params = mapped_info[0]
        A, B = best_params
        A_, B_ = 1/A, -B/A

        size2 = len(curve2.x)
        for rec1 in major_peak_info1:
            top1 = rec1[1]
            mapped_top1 = top1*A_ + B_
            for rec2 in major_peak_info2:
                top2 = rec2[1]
                if abs(mapped_top1 - top2)/size2 < 0.05:
                    adopted_peak_info1.append(rec1)
                    adopted_peak_info2.append(rec2)
                    break

        points1, weights1 = curve1.feature.get_points(adopted_peak_info1)
        points2, weights2 = curve2.feature.get_points(adopted_peak_info2)

        num_points1 = len(points1)
        num_points2 = len(points2)

        diff_num = abs(num_points1 - num_points2)

        if REDUCE_WHEN_INFEASIBLE and diff_num > FEASIBLE_DIFF_NUM:
            """
            not yet usable for 20170426
            """
            print('num_points1=', num_points1, 'num_points2=', num_points2)

            A_, B_ = 1/A, -B/A
            self.logger.warning("reducing points: %d vs. %d." % (num_points1, num_points2))

            def reduce_num_points(scale, shift, more_points, more_weights, less_points, less_weights):
                np1 = len(more_points)
                np2 = len(less_points)
                min_distances = [np.min( np.abs( scale*p+shift - less_points) ) for p in more_points]
                k = np1 - (np2 + FEASIBLE_DIFF_NUM)
                ppp = np.argpartition(min_distances, k)
                adopted_index = ppp[k:]
                print('adopted_index=', adopted_index)
                ret_points = more_points[adopted_index]
                ret_weights = more_weights[adopted_index]
                return ret_points, ret_weights

            if num_points1 > num_points2:
                points1, weights1 = reduce_num_points(A_, B_, points1, weights1, points2, weights2)
                num_points1 = len(points1)
            else:
                points2, weights2 = reduce_num_points(A, B, points2, weights2, points1, weights1)
                num_points2 = len(points2)

            diff_num = num_points1 - num_points2
            self.logger.warning("reduced points: %d vs. %d." % (num_points1, num_points2))

            if False:
                names = ['UV', 'Xray']
                colors = ['blue', 'orange']

                plt.push()
                fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,7))
                ax1, ax2 = axes
                simple_plot(ax1, curve1, names[0], color=colors[0], legend=False)
                simple_plot(ax2, curve2, names[1], color=colors[1], legend=False)
                ax1.plot(points1, curve1.spline(points1), 'o', color='yellow')
                ax2.plot(points2, curve2.spline(points2), 'o', color='yellow')

                mapped_points = A_*points1+B_
                ax2.plot(mapped_points, curve2.spline(mapped_points), 'o', color='cyan')
                fig.tight_layout()
                plt.show()
                plt.pop()

        assert diff_num <= FEASIBLE_DIFF_NUM

        greater_num = max(num_points1, num_points2)
        less_num = min(num_points1, num_points2)

        greater_index = list(range(greater_num))
        full_index1 = list(range(num_points1))
        full_index2 = list(range(num_points2))

        max_score   = None
        max_params  = None
        max_indeces = None
        max_correl  = None
        max_simil   = None
        max_info    = None

        for c in combinations(greater_index, less_num):
            index = list(c)
            if num_points1 >= num_points2:
                index1 = index
                index2 = full_index2
            else:
                index1 = full_index1
                index2 = index
            try:
                correl, simil, params, info = self.evaluate_mapping_features(curve1, points1, weights1, index1, curve2, points2, weights2, index2)
            except AssertionError:
                # invalid combination
                if False:
                    log_exception(self.logger, "evaluate_mapping_features failure: ")
                continue

            score = correl + simil
            if max_score is None or score > max_score:
                max_score   = score
                max_params  = params
                max_indeces = (index1, index2)
                max_correl  = correl
                max_simil   = simil
                max_info    = info

        pre_recog_covered = False
        if max_score is None:
            if self.pre_recog is None:
                assert False
            else:
                best_params, max_score_pair, max_info = self.pre_recog.get_restricted_mapped_info()
                a, b = best_params
                max_params = np.array([b, a])
                x, y, w = max_info
                correl, simil = self.compute_mapped_curve_simimarity(max_params, x, y, w, None, curve1, curve2, features=True, nodebug=True)
                max_score = correl + simil
                # max_score = 2
                max_corel = correl
                max_simil = simil
                pre_recog_covered = True
                self.logger.warning("to cover the failure, feature matching params are replaced with the pre-recognized params.")

        return max_score, max_params, max_indeces, max_correl, max_simil, max_info

    def find_best_mapping_impl(self, curve1, curve2, features, check_adequacy=True, mapped_info=None, debug=False):

        num_peaks_1 = len(curve1.peak_info)
        num_peaks_2 = len(curve2.peak_info)
        assert num_peaks_1 == len(curve1.boundaries) + 1
        assert num_peaks_2 == len(curve2.boundaries) + 1

        debug = False
        # if debug and self.debug:
        if debug:
            def best_debug_plot(title, curve1, curve2, feature_info=None, orig_curves=None):
                from  molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
                reversed_=False
                plt.push()
                fig, axes = plt.subplots( figsize=(21, 8), nrows=1, ncols=2 )
                fig.suptitle(title + ' for ' + get_in_folder(), fontsize=30)
                ax1, ax2 = axes
                if reversed_:
                    c1, c2 = curve2, curve1
                else:
                    c1, c2 = curve1, curve2

                if orig_curves is None:
                    c1_orig = c1
                    c2_orig = c2
                else:
                    if reversed_:
                        c2_orig, c1_orig = orig_curves
                    else:
                        c1_orig, c2_orig = orig_curves

                points1, weights1 = c1_orig.feature.get_points()
                points2, weights2 = c2_orig.feature.get_points()
                ax1.plot(points1, c1_orig.spline(points1), 'o', color='yellow', markersize=20, label='original feature points')
                ax2.plot(points2, c2_orig.spline(points2), 'o', color='yellow', markersize=20, label='original feature points')

                if orig_curves is not None and feature_info is not None:
                    points1, weights1 = c1.feature.get_points()
                    points2, weights2 = c2.feature.get_points()
                    alt_index, alt_params = feature_info
                    if len(points1) < len(points2):
                        index1 = list(range(len(points1)))
                        index2 = alt_index
                    else:
                        index1 = alt_index
                        index2 = list(range(len(points2)))
                    points1_ = points1[index1]
                    points2_ = points2[index2]
                    ax1.plot(points1_, c1.spline(points1_), 'o', color='cyan', markersize=15, label='mapped feature points')
                    ax2.plot(points2_, c2.spline(points2_), 'o', color='cyan', markersize=15, label='mapped feature points')

                    # mapped_points = points_[alt_index]
                    # ax_.plot(mapped_points, c_.spline(mapped_points), color='cyan', markersize=10, label='mapped points')

                simple_plot(ax1, c1, "UV", color='blue', legend=False)
                simple_plot(ax2, c2, "Xray", color='orange', legend=False)

                ax1.legend(fontsize=16)
                ax2.legend(fontsize=16)
                fig.tight_layout()
                fig.subplots_adjust(top=0.85)
                plt.show()
                plt.pop()
            best_debug_plot("find_best_mapping begin", curve1, curve2)

        feature_info = None

        if features:
            try:
                max_score, max_params, max_indeces, max_correl, max_simil, max_info = self.get_best_mapping_features_params(curve1, curve2, mapped_info=mapped_info)
                feature_info = [ max_indeces, max_params ]
                B, A = max_params
                best_params = np.array([A, B])
                self.logger.info("get_best_mapping_features_params has been successful: A=%g, B=%g" % (A, B))
            except Exception as etb:
                self.logger.warning(" " + etb.last_lines())
                log_exception(self.logger, "get_best_mapping_features_params failed: ")
                max_score, best_params, max_indeces, max_correl, max_simil, max_info = 0, None, None, None, None, None
            curve1_, curve2_ = curve1, curve2
        else:
            index1 = list(range(num_peaks_1))
            max_score, max_params, max_indeces, max_correl, max_simil, max_info = self.get_best_mapping_params(curve1, curve2)
            # curve1_, curve2_, best_params = self.reconstruct_curves(curve1, curve2, max_params, max_indeces, max_correl, max_simil, check_adequacy)
            curve1_, curve2_, best_params = self.reconstruct_curves_simple(curve1, curve2, max_params, max_indeces,)

        best_info = [max_score, best_params, max_indeces, max_correl, max_simil]
        mapped_info = MappedInfo([best_params, [max_correl, max_simil], max_info], [self.a_curve, self.x_curve])

        # if self.debug and debug:
        if debug:
            best_debug_plot("find_best_mapping result", curve1_, curve2_, feature_info, orig_curves=[curve1, curve2])

        return curve1_, curve2_, best_info, mapped_info

    def reconstruct_curves_simple(self, curve1, curve2, max_params, max_indeces):
        B, A = max_params
        A_  = 1/A
        B_  = -B/A

        new_peak_top_x1 = []
        new_peak_info1 = []
        new_boundaries1 = []
        new_peak_top_x2 = []
        new_peak_info2 = []
        new_boundaries2 = []

        index1, index2 = max_indeces
        prev_pair = None
        for n, pair in enumerate(zip(index1, index2)):
            m, k = pair

            top_x1 = curve1.peak_top_x[m]
            peak_rec1 = curve1.peak_info[m]
            new_peak_top_x1.append(top_x1)
            new_peak_info1.append(peak_rec1)

            top_x2 = curve2.peak_top_x[k]
            peak_rec2 = curve2.peak_info[k]
            new_peak_top_x2.append(top_x2)
            new_peak_info2.append(peak_rec2)

            if n > 0:
                boundary1, boundary2 = self.find_valley_bottom_pair(curve1, curve2, prev_pair, m, k, A_, B_)
                new_boundaries1.append(boundary1)
                new_boundaries2.append(boundary2)

            prev_pair = pair

        assert len(new_peak_info1) == len(new_boundaries1) + 1
        assert len(new_peak_info2) == len(new_boundaries2) + 1
        assert len(new_peak_info1) == len(new_peak_info2)

        curve1_ =  ProxyCurve( curve1, new_peak_top_x1, new_peak_info1, new_boundaries1, [], [] )
        curve2_ =  ProxyCurve( curve2, new_peak_top_x2, new_peak_info2, new_boundaries2, [], [] )

        best_params = np.array([A, B])
        return curve1_, curve2_, best_params

    def find_valley_bottom_pair(self, curve1, curve2, prev_pair, m, k, A_, B_):
        prev_m, prev_k = prev_pair
        bc1_list = curve1.boundaries[prev_m:m]
        bc2_list = curve2.boundaries[prev_k:k]

        # select the best matching bottom pair
        min_dist = None
        min_j = None
        min_i = None
        for j, bc1 in enumerate(bc1_list):
            for i, bc2 in enumerate(bc2_list):
                dist = abs(bc1*A_ + B_ - bc2 )
                if min_dist is None or dist < min_dist:
                    min_dist = dist
                    min_j = j
                    min_i = i

        return bc1_list[min_j], bc2_list[min_i]

    def add_corresponding_peaks(self, curve1, curve2, A_, B_, debug_text=""):
        debug = True
        if debug:
            from matplotlib.patches import Rectangle
            from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
            print('A_, B_=', A_, B_)
            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 7))
            ax1, ax2 = axes
            fig.suptitle("add_corresponding_peaks entry: " + debug_text, fontsize=20)
            simple_plot(ax1, curve1, "curve1", color="blue")
            simple_plot(ax2, curve2, "curve2", color="orange")

            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(ymin, ymax)
            allow = len(curve2.x) * 0.05
            for top_x in curve2.peak_top_x:
                f = top_x - allow
                t = top_x + allow
                p = Rectangle(
                        (f, ymin),  # (x,y)
                        t - f,   # width
                        ymax - ymin,    # height
                        facecolor   = 'yellow',
                        alpha       = 0.3,
                    )
                ax2.add_patch(p)

            xmin, xmax = ax2.get_xlim()
            ax2t = ax2.twinx()
            ax2t.grid(False)
            ax2t.set_xlim(xmin, xmax)
            mapped_x = A_ * curve1.x  + B_
            ax2t.plot(mapped_x, curve1.y, ':', color='blue', label='mapped data')
            for top_x in curve1.peak_top_x:
                mx = A_ * top_x + B_
                my = curve1.spline(top_x)
                ax2t.plot(mx, my, 'o', color='yellow', label='mapped peak tops')

            ax2t.legend(bbox_to_anchor=(1, 0.9), loc='upper right')
            fig.tight_layout()
            plt.show()
            plt.pop()

        new_peak_top_x2 = []
        new_peak_info2 = []
        new_boundaries2 = []

        size2 = len(curve2.x)
        peaks2 = []
        start = 0
        stop = None
        last_peak_rec2 = None
        for m, top_x1 in enumerate(curve1.peak_top_x):
            mapped_top_x = A_*top_x1 + B_

            bottom_added = False
            if m > 0 and len(new_peak_top_x2) > 0:
                try:
                    bottom_x1 = curve1.boundaries[m-1]
                    mapped_bottom_x = A_*bottom_x1 + B_
                    bottom_x2 = self.find_counter_part_bottom(curve2, mapped_bottom_x, debug)
                    new_boundaries2.append(bottom_x2)
                    bottom_added = True
                except AssertionError:
                    pass

            try:
                mapped_peak_rec = [int(A_*i + B_ + 0.5) for i in curve1.peak_info[m]]
                top_x2, peak_rec2 = self.find_counter_part_peak(curve2, mapped_top_x, mapped_peak_rec, debug)
                new_peak_top_x2.append(top_x2)
                if last_peak_rec2 is not None:
                    if last_peak_rec2[2] >= peak_rec2[0]:
                        b = new_boundaries2[-1]
                        r = peak_rec2[0]
                        peak_rec2[0] = int((last_peak_rec2[2] + b)/2)
                        self.logger.warning("the left end modified from %d to %d in the %d-th range in add_corresponding_peaks", m, r, peak_rec2[0])
                new_peak_info2.append(peak_rec2)
                last_peak_rec2 = peak_rec2
            except AssertionError:
                warnlog_exception(self.logger, "add_corresponding_peaks: try peak")
                if bottom_added:
                    new_boundaries2.pop(-1)

            if debug:
                print("--------------------", [m], "new_peak_info2=", new_peak_info2)

        # assert len(new_peak_top_x2) == len(new_boundaries2) + 1

        if len(new_peak_top_x2) == len(new_boundaries2) + 1:
            curve2_ =  ProxyCurve( curve2, new_peak_top_x2, new_peak_info2, new_boundaries2, [], [] )
        else:
            curve2_ = curve2

        if debug:
            print('len(new_peak_top_x2)=', len(new_peak_top_x2))
            print('len(new_boundaries2)=', len(new_boundaries2))
            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 7))
            ax1, ax2 = axes
            fig.suptitle("add_corresponding_peaks exit: " + debug_text)
            simple_plot(ax1, curve1, "curve1")
            simple_plot(ax2, curve2_, "curve2_")
            fig.tight_layout()
            plt.show()
            plt.pop()

        return curve2_

    def find_the_nearest_index(self, curve, vecx, px):
        """
        must choose the nearest since diff_ratio < 0.05 can detect multiple candidates
        """
        pr_width = curve.get_peak_region_width()

        min_ratio = None
        min_k = None
        for k, x_ in enumerate(vecx):
            diff_ratio = abs(px - x_)/pr_width
            if diff_ratio < 0.05:
                if min_ratio is None or diff_ratio < min_ratio:
                    min_k = k
                    min_ratio = diff_ratio

    def find_counter_part_peak(self, curve, ref_peak_x, ref_peak_rec, debug=False):
        k = self.find_the_nearest_index(curve, curve.peak_top_x, ref_peak_x)
        if k is not None:
            if debug:
                print('find_counter_part_peak: found', k)
            return curve.peak_top_x[k], curve.peak_info[k]

        delta_x = len(curve.x)*0.05
        start = int(ref_peak_x - delta_x)
        stop  = int(ref_peak_x + delta_x)

        n = find_rotated_extreme_arg(curve, start, stop, +1)

        # avoid minor peaks such as the point near 650 in 20170304 UV
        # default delta_x is too small for the 1st minor peak in 20160227
        assert curve.locally_tall_enough_evenif_rotated(n, delta_x=delta_x)

        if debug:
            print('-------------------- find_counter_part_peak: found rotated', n)

        return n, [ref_peak_rec[0], n, ref_peak_rec[2]]

    def find_counter_part_bottom(self, curve, ref_bottom_x, debug=False):
        k = self.find_the_nearest_index(curve, curve.boundaries, ref_bottom_x)
        if k is not None:
            return curve.boundaries[k]

        delta_x = len(curve.x)*0.05
        start = int(ref_bottom_x - delta_x)
        stop  = int(ref_bottom_x + delta_x)

        n = find_rotated_extreme_arg(curve, start, stop, -1)

        if debug:
            print('-------------------- find_counter_part_bottom: found rotated', n)

        return n

    def show_mapping_proof(self):
        a_curve_, x_curve_  = self.get_mapped_curves()
        mapping_proof("mapping proof", a_curve_, x_curve_, self.best_info[1])

    def make_better_curve(self, curve1, curve2, mapper, mapping, debug_text, debug=False):
        nomatch = mapper[mapper < 0]
        if len(nomatch) == 0:
            return curve1

        A, B = mapping

        new_peak_top_x = []
        new_peak_info = []
        new_boundaries = []

        max_y = curve1.max_y
        ptx1 = curve1.peak_top_x
        pif1 = curve1.peak_info
        bdy1 = curve1.boundaries
        ptx_ = curve2.peak_top_x*A + B
        bry_ = np.array(curve2.boundaries)*A + B

        if debug:
            print("mapper=", mapper)
            if debug_text.find("a_curve") >= 0:
                curve1_id = "UV"
                curve2_id = "XR"
            else:
                curve1_id = "XR"
                curve2_id = "UV"
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("%s entry" % debug_text)
                ax1.set_title(curve1_id + " Elution")
                ax2.set_title(curve2_id + " Elution")
                simple_plot(ax1, curve1, legend=False)

                pty_ = curve1.spline(ptx_)
                ax1.plot(ptx_, pty_, "o", color="cyan", markersize=20, alpha=0.2, label="mapped peak")
                ax1.legend()

                simple_plot(ax2, curve2)
                fig.tight_layout()
                plt.show()

        ins = 0
        last_peak_info = None
        def adjusted_peak_info(temp_info):
            if last_peak_info is None or last_peak_info[2] < temp_info[0]:
                return temp_info

            ret_info = [last_peak_info[2]+1, *temp_info[1:]]
            assert ret_info[0] < ret_info[1]

            return ret_info

        try:
            for k, j in enumerate(mapper):
                if j >= 0:
                    continue
                x_ = ptx_[k]
                i = bisect_right(ptx1, x_)
                print([k], "i=", i)
                n = 0
                for p in range(ins, i):
                    new_peak_top_x.append(ptx1[p])
                    peak_info = adjusted_peak_info(pif1[p])
                    new_peak_info.append(peak_info)
                    last_peak_info = peak_info
                    if p > 0:
                        new_boundaries.append(bdy1[p-1])
                    n += 1
                ins += n
                top_x = ptx_[i]

                height_ratio = curve1.spline(top_x)/max_y
                # print([k], "-------------------- height_ratio=", height_ratio)
                if height_ratio < MINIMUM_HEIGHT_RATIO:
                    self.logger.info("peak with height_ratio=%.3g < %.3g will be discarded.", height_ratio, MINIMUM_HEIGHT_RATIO)
                    # eventually resorting to make_thinner_curve
                    assert False

                new_peak_top_x.append(top_x)
                peak_info = adjusted_peak_info([int(round(p*A + B)) for p in curve2.peak_info[i]])
                new_peak_info.append(peak_info)
                last_peak_info = last_peak_info
                if len(bry_) > 0:
                    new_boundaries.append(bry_[max(0, i-1)])    # task: prove max(0, i-1) is correct
        except AssertionError as exc:
            if debug:
                mx_ = curve2.x*A + B
                scale = curve1.max_y/curve2.max_y
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("adjusted_peak_info AssertionError")
                    simple_plot(ax, curve1, legend=False)
                    ax.plot(mx_, scale*curve2.y, ":", label="mapped")
                    ax.plot(ptx_, scale*curve2.spline(curve2.peak_top_x), "o", markersize=20, color="cyan", alpha=0.2)
                    ax.legend()
                    fig.tight_layout()
                    plt.show()
            raise exc

        for p in range(ins, len(ptx1)):
            new_peak_top_x.append(ptx1[p])
            new_peak_info.append(pif1[p])
            if p > 0:
                new_boundaries.append(bdy1[p-1])

        curve1_ =  ProxyCurve( curve1, new_peak_top_x, new_peak_info, new_boundaries, [], [] )

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("%s result" % debug_text)
                simple_plot(ax1, curve1_, legend=False)

                pty_ = curve1_.spline(ptx_)
                ax1.plot(ptx_, pty_, "o", color="cyan", markersize=20, alpha=0.2, label="mapped peak")
                ax1.legend()

                simple_plot(ax2, curve2)
                fig.tight_layout()
                plt.show()

        return curve1_

    def make_thinner_curve(self, curve, mapper):
        nomatch = mapper[mapper < 0]
        if len(nomatch) == 0:
            return curve

        new_peak_top_x = []
        new_peak_info = []
        new_boundaries = []

        ptx = curve.peak_top_x
        pif = curve.peak_info
        bdy = curve.boundaries

        n = 0
        for k, j in enumerate(mapper):
            if j < 0:
                # primary_peak must not be lost
                if k == curve.primary_peak_no:
                    raise AssertionError("primary peak lost")
                continue

            new_peak_top_x.append(ptx[k])
            new_peak_info.append(pif[k])
            if n > 0 and k > 0:
                new_boundaries.append(bdy[k-1])
            n += 1

        curve_ =  ProxyCurve( curve, new_peak_top_x, new_peak_info, new_boundaries, [], [] )
        return curve_

    def can_be_deemed_single_peak(self, curves):
        if len(curves[0].peak_info) == 1:
            return True

        num_peaks = []
        for curve in curves:
            n = 0
            max_y = curve.max_y
            for info in curve.peak_info:
                ratio = curve.spline(info[1])/max_y
                if ratio > DEEM_SINGLE_PEAK_RATIO:
                    n += 1
            num_peaks.append(n)
        return np.min(num_peaks) == 1

    def get_imroved_mapping(self, fallback=False, debug=False):
        from  molass_legacy.DataStructure.LPM import get_corrected   # moved due to ImportError: ... (most likely due to a circular import)

        if debug:
            from  molass_legacy.KekLib.DebugUtils import show_call_stack
            show_call_stack("----: ", indented_only=True)            

        curves = self.mapped_curves
        A, B = self.best_info[1]

        if not self.can_be_deemed_single_peak(curves):
            return A, B

        max_simil = self.best_info[4]
        if len(curves[0].peak_info) == 1:
            uv_px = curves[0].peak_info[0][1]
            mp_px = A * curves[1].peak_info[0][1] + B
            dx = uv_px - mp_px
            self.logger.info("an adjustment %.3g of peak top deviation will be applied.", dx)
        else:
            dx = 0

        B += dx
        if max_simil < ACCEPTABLE_SIMILARITY:
            in_fallback = " in fallback" if fallback else ""
            self.logger.info("trying to improve to mapping due to poor similarity %.3g < %.3g%s.", max_simil, ACCEPTABLE_SIMILARITY, in_fallback)
        else:
            return A, B

        x1 = curves[0].x
        y1 = get_corrected(curves[0].y)
        i1 = curves[0].primary_peak_i
        x2 = curves[1].x
        y2 = get_corrected(curves[1].y)
        i2 = curves[1].primary_peak_i

        spline = UnivariateSpline(x1, y1, s=0, ext=3)
        scale = y1[i1]/y2[i2]

        def objective(p):
            a, b = p
            x = a*x2 + b
            return np.sum((spline(x) - scale*y2)**2)

        ret = minimize(objective, (A, B), method="Nelder-Mead")
        a, b = ret.x

        mapped_ends = [ex*a + b for ex in x2[[0, -1]]]
        ratio = (mapped_ends[1] - mapped_ends[0])/(x1[-1] - x1[0])
        if abs(ratio - 1) > ACCEPTABLE_ENDS_RATIO_DEV:
            # as in 20200123_4
            self.logger.info("mapping (%.3g, %.3g) is discarded due to ratio=%.3g", a, b, ratio)
            a, b = A, B

        if debug:
            print("fallback=", fallback)
            spline2 = UnivariateSpline(x2, y2, s=0, ext=3)
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("get_imroved_mapping")
                ax1.set_title("Uncertain Mapping")
                ax1.plot(x1, y1)
                for info in curves[0].peak_info:
                    px = info[1]
                    ax1.plot(px, spline(px), "o", color="red")
                ax1.plot(A*x2+B, scale*y2, ":")
                for info in curves[1].peak_info:
                    px = info[1]
                    ux = A*px + B
                    ax1.plot(ux, scale*spline2(px), "o", color="cyan", markersize=20, alpha=0.3)

                def draw_slice(ax, slice_):
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim(ymin, ymax)
                    for j in [slice_.start, slice_.stop]:
                        ax.plot([j, j], [ymin, ymax], ":", color="yellow")

                draw_slice(ax1, curves[0].peak_slice)

                ax2.set_title("Improved Mapping")
                ax2.plot(x1, y1)
                for info in curves[0].peak_info:
                    px = info[1]
                    ax2.plot(px, spline(px), "o", color="red")
                ax2.plot(a*x2+b, scale*y2, ":")
                for info in curves[1].peak_info:
                    px = info[1]
                    ux = a*px + b
                    ax2.plot(ux, scale*spline2(px), "o", color="cyan", markersize=20, alpha=0.3)

                draw_slice(ax2, curves[0].peak_slice)

                fig.tight_layout()
                plt.show()

        return a, b

    def compare_primary_peak_mapping_params(self, curve1, curve2, mapping, debug=False):
        px1 = curve1.primary_peak_x
        px2 = curve2.primary_peak_x
        scale = curve1.spline(px1)/curve2.spline(px2)
        uv_x = curve1.x
        x2 = curve2.x

        def objective(p, add_no_mapping=False, debug=False):
            a, b, w = p
            if w <= 1e-3:
                return np.inf

            x1 = a*x2 + b
            compare_interval = np.logical_and(px1 - w <= x1, x1 <= px1 + w)
            x1_ = x1[compare_interval]
            x2_ = x2[compare_interval]
            dev = np.sum((curve1.spline(x1_) - scale*curve2.spline(x2_))**2)

            if add_no_mapping:
                # add penalty from region with no mapping as in OA_Ald with different primary peaks
                if x1[0] < uv_x[0]:
                    extend_L = np.arange(x1[0], uv_x[0], 1)   # uv_x[1] - uv_x[0] == 1
                else:
                    extend_L = []
                if uv_x[-1] < x1[-1]:
                    extend_R = np.arange(uv_x[-1]+1, x1[-1]+1, 1)
                else:
                    extend_R = []
                ex_uv_x = np.concatenate([extend_L, uv_x, extend_R])
                penalty = np.sum(curve1.spline(ex_uv_x[ex_uv_x < x1[0]])**2) + np.sum(curve1.spline(ex_uv_x[ex_uv_x > x1[-1]])**2)
            else:
                penalty = 0

            if debug:
                print("dev/w=", dev/w)
                print("penalty=", penalty)
                with plt.Dp():
                    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                    fig.suptitle("objective debug")

                    simple_plot(ax1, curve1, legend=False)
                    simple_plot(ax2, curve2, legend=False)

                    y1_ = curve1.spline(x1_)
                    y2_ = curve2.spline(x2_)
                    ax1.plot(x1_, y1_, ":", label="compare_interval")
                    ax2.plot(x2_, y2_, ":", label="compare_interval")
                    ax1.fill_between(x1_, y1_, scale*y2_, color='pink', alpha=0.2)

                    for ax in ax1, ax2:
                        ax.legend()

                    fig.tight_layout()
                    plt.show()

            return dev/w + penalty

        a_init = len(curve1.x)/len(curve2.x)
        b_init = px1 - a_init*px2
        pinfo = curve1.peak_info[curve1.primary_peak_no]
        w_init = (pinfo[2] - pinfo[0])/2
        ret = minimize(objective, (a_init, b_init, w_init))
        a, b, w = ret.x
        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("compare_primary_peak_mapping_params")
                simple_plot(ax, curve1)
                x1 = a*x2 + b
                ax.plot(x1, scale*curve2.y, ":")
                fig.tight_layout()
                plt.show()

        W = len(curve1.x)
        primary_score = objective((a, b, W), add_no_mapping=True)
        A, B = mapping
        compare_score = objective((A, B, W), add_no_mapping=True)
        if primary_score >= compare_score:
            self.logger.info("adopt normal peak mapping with score comparison %.3g <= %.3g", compare_score, primary_score)
            a, b = A, B
        else:
            self.logger.info("adopt primary peak mapping with score comparison %.3g < %.3g", primary_score, compare_score)

        return a, b



"""
    Estimators.EdmEstimatorDevel.py
"""
import numpy as np
import matplotlib.pyplot as plt
from .EghEstimator import EghEstimator

class MappingProxy:
    def __init__(self, a, b):
        self.slope = a
        self.intercept = b

class SsdProxy:
    def __init__(self, mapping):
        self.xr = None
        self.uv = None
        self.mapping = mapping

    def get_mapping(self):
        return self.mapping

class EdmEstimatorDevel(EghEstimator):
    def __init__(self, editor, n_components):
        self.n_components = n_components
        EghEstimator.__init__(self, editor)

    def estimate_params(self, debug=False):
        from molass.LowRank.Decomposition import Decomposition
        from molass.LowRank.ComponentCurve import ComponentCurve
        from molass.PlotUtils.DecompositionPlot import plot_elution_curve

        init_xr_params, init_xr_baseparams, temp_rgs, init_mapping, init_uv_heights, init_uv_baseparams, init_mappable_range, seccol_params = self.estimate_egh_params()
        print("Initial XR Params:", init_xr_params)
        uv_curve, xr_curve = self.ecurves
        xr_x = xr_curve.x

        egh_xr_ccurves = []
        for i, param in enumerate(init_xr_params):
            print(f"  Component {i}: {param}")
            egh_xr_ccurves.append(ComponentCurve(xr_x, param))

        uv_x = uv_curve.x
        a, b = init_mapping
        egh_uv_ccurves = []
        for i, (h, params) in enumerate(zip(init_uv_heights, init_xr_params)):
            tR = a * params[1] + b
            sigma = a * params[2]
            tau = a * params[3]
            uv_params = np.array([h, tR, sigma, tau])
            egh_uv_ccurves.append(ComponentCurve(uv_x, uv_params))

        ssd = SsdProxy(MappingProxy(*init_mapping))
        xr_icurve = xr_curve    # compatible?
        uv_icurve = uv_curve    # compatible?

        decomposition = Decomposition(ssd, xr_icurve, egh_xr_ccurves, uv_icurve, egh_uv_ccurves)
        if debug:
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10,4))
            fig.suptitle("EGH Initial Parameters by EdmEstimatorDevel")
            plot_elution_curve(ax1, uv_curve, egh_uv_ccurves)
            plot_elution_curve(ax2, xr_curve, egh_xr_ccurves)
            plt.show()

        edm_decomposition = decomposition.optimize_with_model('EDM')
        edm_uv_ccurves = edm_decomposition.uv_ccurves
        edm_xr_ccurves = edm_decomposition.xr_ccurves
        if debug:
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10,4))
            fig.suptitle("EDM Initial Parameters by EdmEstimatorDevel")
            plot_elution_curve(ax1, uv_curve, edm_uv_ccurves)
            plot_elution_curve(ax2, xr_curve, edm_xr_ccurves)
            plt.show()
 
        xr_params_list = []
        for ccurve in edm_xr_ccurves:
            xr_params_list.append(ccurve.params)
        xr_params = np.array(xr_params_list)
        Tz = np.average(xr_params[:,0])
        nc = self.n_components - 1
        uv_w = np.array([uv_curve.max_y/xr_curve.max_y] * nc)
        init_params = np.concatenate([xr_params.flatten(), init_xr_baseparams, temp_rgs, init_mapping, uv_w, init_uv_baseparams, init_mappable_range, [Tz]])
        return init_params


"""
KekLib.IpyUtils.py - Utility functions for interactive user prompts in Jupyter notebooks.
"""
import ipywidgets as widgets
from IPython.display import display, clear_output

def ask_user(question, callback=None, output_widget=None):
    """
    Display a Yes/No dialog in a dedicated Output widget.

    Parameters:
    -----------
    question : str
        The question to present to the user.
    callback : function, optional
        A function to call with the user's response (True for Yes, False for No).
    output_widget : ipywidgets.Output, optional
        An existing Output widget to use for displaying the dialog. If None, a new one is created.
        
    Returns:
    --------
    output_widget : ipywidgets.Output
    """
    if output_widget is None:
        output_widget = widgets.Output()
        display(output_widget)  # Only display if not already displayed

    with output_widget:
        clear_output(wait=True)
        label = widgets.Label(question)
        btn_yes = widgets.Button(description="Yes", button_style='success')
        btn_no = widgets.Button(description="No", button_style='danger')
        button_box = widgets.HBox([btn_yes, btn_no])
        display(widgets.VBox([label, button_box]))

        def on_yes(b):
            with output_widget:
                clear_output(wait=True)
                print("You selected: Yes")
            if callback:
                callback(True)

        def on_no(b):
            with output_widget:
                clear_output(wait=True)
                print("You selected: No")
            if callback:
                callback(False)

        btn_yes.on_click(on_yes)
        btn_no.on_click(on_no)

    return output_widget


"""
Optimizer.JobStatePlot.py
Job status plot for optimization GUI.
"""
import numpy as np
import matplotlib.pyplot as plt

def draw_suptitle(self):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
    from molass_legacy.KekLib.BasicUtils import ordinal_str
    from molass_legacy.Optimizer.OptimizerUtils import get_model_name, get_method_name
    job_name = "%03d" % (self.num_trials,)
    in_folder = get_in_folder()
    model_name = get_model_name(self.func_code)
    text = "Job %s State at %s local minimum on %s with model=%s method=%s" % (
        job_name, ordinal_str(self.curr_index), in_folder, model_name, get_method_name())
    if self.suptitle is None or True:
        self.suptitle = self.fig.suptitle(text, fontsize=20)
    else:
        self.suptitle.set_text(text)

def plot_job_state(self, params, plot_info=None, niter=20):
    from matplotlib.gridspec import GridSpec
    import seaborn
    seaborn.set_theme()
    from importlib import reload
    import molass_legacy.Optimizer.ProgressChart
    reload(molass_legacy.Optimizer.ProgressChart)
    from molass_legacy.Optimizer.ProgressChart import draw_progress

    self.fig = fig = plt.figure(figsize=(18, 9))
    gs = GridSpec(33, 15, wspace=1.3, hspace=1.0)
    axes = []
    for j in range(3):
        j_ = j*5
        ax = fig.add_subplot(gs[0:16,j_:j_+5])
        axes.append(ax)

    axt = axes[1].twinx()
    axt.grid(False)
    axes.append(axt)
    self.axes = axes
    self.prog_ax = fig.add_subplot(gs[17:21,2:])
    peak_ax = fig.add_subplot(gs[21:25,2:])
    rg_ax = fig.add_subplot(gs[25:29,2:])
    map_ax = fig.add_subplot(gs[29:33,2:])
    self.prog_axes = [self.prog_ax, peak_ax, rg_ax, map_ax]
    for ax in self.prog_axes[0:3]:
        ax.set_xticklabels([])
    self.prog_title_axes = [fig.add_subplot(gs[17+i*4:21+i*4,0:2]) for i in range(0,4)]
    prog_titles = ["Function SV", "Peak Top Positions", "Rg Values", "Mapped Range"]
    for ax, title in zip(self.prog_title_axes, prog_titles):
        ax.set_axis_off()
        ax.text(-0.3, 0.5, title, fontsize=16)

    draw_suptitle(self)
    plot_objective_func(self.optimizer, params, axis_info=(self.fig, self.axes))

    if plot_info is not None:
        draw_progress(self, plot_info, niter=niter)

def plot_objective_func(optimizer, params, axis_info=None):
    from .FvScoreConverter import convert_score
    fv_ = optimizer.objective_func(params)
    sv = convert_score(fv_)

    if axis_info is None:
        fig, axes = plt.subplots(ncols=3, figsize=(18,4.5))
        ax1, ax2, ax3 = axes
        axt = ax2.twinx()
        axt.grid(False)
        axis_info = (fig, (*axes, axt))
    else:
        fig, axes = axis_info
        ax1, ax2, ax3 = axes[:3]

    ax1.set_title("UV Decomposition", fontsize=16)
    ax2.set_title("Xray Decomposition", fontsize=16)
    ax3.set_title("Objective Function Scores in SV=%.3g" % sv, fontsize=16)
    optimizer.objective_func(params, plot=True, axis_info=axis_info)


"""
    Optimizer.OptimizerMain.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import sys
import os
import getopt
import numpy as np
from molass_legacy.Baseline.BaselineUtils import create_xr_baseline_object
from .FuncImporter import import_objective_function

def main_driver():
    optlist, args = getopt.getopt(sys.argv[1:], 'c:w:f:n:i:b:d:m:s:r:t:p:T:M:S:L:X:O:')
    print(optlist, args)
    optdict = dict(optlist)
    if optdict.get('-r'):
        main_impl(optdict, optlist)
    else:
        main_tk(optdict, optlist)

def main_tk(optdict, optlist):
    from molass_legacy.KekLib.TkUtils import get_tk_root
    root = get_tk_root()
    def run_main():
        main_impl(optdict, optlist)
        root.quit()
    root.after(0, run_main)
    root.mainloop()
    root.destroy()

def main_impl(optdict, optlist):
    import sys
    import os
    import numpy as np
    from molass_legacy.KekLib.ChangeableLogger import Logger
    from molass_legacy import get_version
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Optimizer.OptimizerSettings import OptimizerSettings
    from molass_legacy.Optimizer.OptimizerMain import optimizer_main
    from molass_legacy.Optimizer.TheUtils import get_analysis_folder_from_work_folder
    from molass_legacy.Optimizer.SettingsSerializer import unserialize_for_optimizer

    work_folder = optdict['-w']
    os.chdir(work_folder)
    work_folder = os.getcwd()   # to get absolute path

    nnn = int(work_folder[-3:])

    log_file = "optimizer.log"
    logger = Logger(log_file)

    try:
        logger.info(get_version(toml_only=True) + " (development version)")
    except Exception as e:
        logger.info(get_version())
    logger.info("Optimizer started in %s", work_folder)
    python_syspath = os.environ.get('MOLASS_PYTHONPATH')
    logger.info("MOLASS_PYTHONPATH=%s", str(python_syspath))

    analysis_folder = get_analysis_folder_from_work_folder(work_folder)
    logger.info("xxx sys.argv[1:]: %s", sys.argv[1:])
    logger.info("work_folder: %s", work_folder)
    logger.info("analysis_folder inferred as %s", analysis_folder)
    set_setting("analysis_folder", analysis_folder)
    optimizer_folder = os.path.join(analysis_folder, "optimized")
    set_setting("optimizer_folder", optimizer_folder)   # optimizer_folder will be referenced in DataTreatment.load()

    in_folder = optdict['-f']
    set_setting("in_folder", in_folder)     # required in the devel mode of SecTheory.ColumnTypes.py

    settings = OptimizerSettings()
    settings.load(optimizer_folder=optimizer_folder)    # this should restore required temporary settings
    logger.info("optimizer settings restored as %s", str(settings))

    try:
        class_code = optdict['-c']
        n_components = int(optdict['-n'])
        init_params_txt = optdict['-i']
        init_params = np.loadtxt(init_params_txt)
        bounds_txt = optdict['-b']
        if os.path.exists(bounds_txt):
            real_bounds = np.loadtxt(bounds_txt)
        else:
            real_bounds = None
        drift_type = optdict['-d']
        niter = int(optdict['-m'])
        seed = int(optdict['-s'])
        trimming_txt = optdict['-r']
        sleep_seconds = optdict.get('-t')
        legacy = optdict.get('-L') == 'legacy'

        unserialize_for_optimizer(optdict.get('-p'))    # "poresize_bounds", "t0_upper_bound"

        test_pattern = optdict.get('-T')
        if test_pattern != "None":
            set_setting("test_pattern", test_pattern)

        with open("in_data_info.txt", "w") as fh:
            fh.write("in_folder=%s\n" % in_folder)

        callback_txt = "callback.txt"
        with open(callback_txt, "w") as fh:
            pass

        with open("pid.txt", "w") as fh:
            fh.write("pid=%d\n" % os.getpid())

        with open("seed.txt", "w") as fh:
            fh.write("seed=%d\n" % seed)

        shm_name = optdict.get('-M')
        if shm_name is None or shm_name  == "None":
            shared_memory = None
        else:
            from molass_legacy.Optimizer.NpSharedMemory import get_shm_proxy
            shared_memory = get_shm_proxy(shm_name)

        solver = optdict.get('-S')
        xr_only = optdict.get('-X') == '1'
        MOLASS_OPTIMIZER_TEST = os.environ.get("MOLASS_OPTIMIZER_TEST")
        logger.info("MOLASS_OPTIMIZER_TEST=%s", str(MOLASS_OPTIMIZER_TEST))
        optimizer_test = (optdict.get('-O') == '1' or MOLASS_OPTIMIZER_TEST == '1')   # workaround for -O not passed issue

        if sleep_seconds is None:
            logger.info("optimizer started with class_code=%s, optlist=%s, shared_memory=%s, xr_only=%s, optimizer_test=%s",
                        class_code, str(optlist), shm_name, xr_only, optimizer_test)
            optimizer_main(in_folder,
                    trimming_txt=trimming_txt,
                    n_components=n_components,
                    solver=solver,
                    drift_type=drift_type,
                    init_params=init_params,
                    real_bounds=real_bounds,
                    niter=niter,
                    seed=seed,
                    class_code=class_code,
                    shared_memory=shared_memory,
                    nnn=nnn,
                    legacy=legacy,
                    xr_only=xr_only,
                    optimizer_test=optimizer_test,
                    debug=False,
                    )

        else:
            from time import sleep
            logger.info("dummy started with niter=%d, seed=%s", niter, str(seed))
            for k in range(int(sleep_seconds)):
                if k % 3 == 0:
                    with open(callback_txt, "a") as fh:
                        fh.write(str([k])+"\n")
                sleep(1)
        if shared_memory is not None:
            shared_memory.close()
            logger.info("shared_memory closed")
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(logger, "main_impl failed: ", n=10)
        exit(-1)

def create_optimizer_from_job(in_folder=None, n_components=None, class_code=None, trimming_txt=None, shared_memory=None,
                              need_settings=False):
    if need_settings:
        pass

    from molass.Bridge.OptimizerInput import OptimizerInput
    fullopt_input = OptimizerInput(in_folder=in_folder, trimming_txt=trimming_txt, legacy=True)
    dsets = fullopt_input.get_dsets()
    x_shifts_file = trimming_txt.replace('trimming.txt', 'x_shifts.txt')
    if os.path.exists(x_shifts_file):
        x_shifts = np.loadtxt(x_shifts_file, dtype=int)
        dsets.apply_x_shifts(x_shifts)
    fullopt_class = import_objective_function(class_code)
    uv_base_curve = fullopt_input.get_base_curve()      # uv_base_curve comes from FullOptInput.get_sd_from_folder()
    xr_base_curve = create_xr_baseline_object()
    qvector, wvector = fullopt_input.get_spectral_vectors()
    optimizer = fullopt_class(dsets, n_components,
                uv_base_curve=uv_base_curve,
                xr_base_curve=xr_base_curve,
                qvector=qvector,
                wvector=wvector,
                shared_memory=shared_memory)
    return optimizer

def optimizer_main(in_folder, trimming_txt=None, n_components=3,
                   solver=None,
                   init_params=None, real_bounds=None,
                   drift_type=None, niter=100, seed=None,
                   callback=True, class_code='F0000', shared_memory=None,
                   nnn=0,
                   legacy=True,
                   xr_only=False,
                   optimizer_test=False,
                   debug=True):

    optimizer = create_optimizer_from_job(in_folder=in_folder,
                                          n_components=n_components,
                                          class_code=class_code,
                                          trimming_txt=trimming_txt,
                                          shared_memory=shared_memory)
    optimizer.set_xr_only(xr_only)

    if optimizer_test:
        from molass_legacy.Optimizer.Compatibility import test_optimizer_compatibility
        test_optimizer_compatibility(optimizer, init_params)
        return

    if seed is None:
        seed = np.random.randint(100000, 999999)
    strategy = optimizer.get_strategy()
    if strategy.trust_initial_baseline():
        baseline_fixed = True
        from molass_legacy.Optimizer.FixedBaselineOptimizer import FixedBaselineOptimizer
        fb_optimizer = FixedBaselineOptimizer(optimizer)
        result = fb_optimizer.solve(init_params, real_bounds=real_bounds, niter=niter, seed=seed, debug=debug)
    else:
        baseline_fixed = False
        if strategy.baseline_first():
            if nnn == 0:
                from molass_legacy.Optimizer.BaselineOptimizer import BaselineOptimizer
                baseline_optimizer = BaselineOptimizer(optimizer)
                baseline_indeces = optimizer.get_baseline_indeces()
                result = baseline_optimizer.solve(init_params, baseline_indeces)
            else:
                baseline_fixed = True
                result = optimizer.solve(init_params, real_bounds=real_bounds, niter=niter, seed=seed, callback=callback, method=solver,
                                        baseline_fixed=baseline_fixed, debug=debug)
        else:
            if strategy.is_strategic(nnn):
                from molass_legacy.Optimizer.StrategicOptimizer import StrategicOptimizer
                temp_params = init_params
                indeces_list = strategy.get_indeces_list(nnn)
                assert len(indeces_list) == 1   # for now
                open_mode = "w"                 # "callback.txt" open mode
                for indeces in indeces_list:
                    strategic_optimizer = StrategicOptimizer(optimizer, indeces)
                    # task: add method option
                    result = strategic_optimizer.solve(temp_params, real_bounds=real_bounds, niter=niter, seed=seed, open_mode=open_mode, debug=debug)
                    temp_params = result.x
                    open_mode = "a"
            else:
                result = optimizer.solve(init_params, real_bounds=real_bounds, niter=niter, seed=seed, callback=callback, method=solver, debug=debug)
    fig_info = [in_folder, None, result]
    if debug:
        optimizer.objective_func(result.x, plot=True, fig_info=fig_info)



"""
    optimizer-dummy.py
"""

def main():
    import os
    import sys
    import getopt
    this_dir = os.path.dirname( os.path.abspath( __file__ ) )
    root_dir = os.path.dirname(os.path.dirname( this_dir ))
    sys.path.insert(0, root_dir)

    optlist, args = getopt.getopt(sys.argv[1:], 'c:w:f:n:i:b:d:m:s:r:t:p:T:M:S:')
    print(optlist, args)
    optdict = dict(optlist)

    from molass_legacy.KekLib.ChangeableLogger import Logger
    from molass_legacy._MOLASS.Version import get_version_string

    work_folder = optdict['-w']
    os.chdir(work_folder)
    work_folder = os.getcwd()   # to get absolute path

    nnn = int(work_folder[-3:])

    log_file = "optimizer.log"
    logger = Logger(log_file)

    logger.info(get_version_string(with_date=True))

    with open("pid.txt", "w") as fh:
        fh.write("pid=%d\n" % os.getpid())

    import time
    time.sleep(60)  # Dummy wait to simulate optimization process

if __name__ == '__main__':
    main()


"""
Optimizer.Compatibility.py
"""
import os
import shutil
import matplotlib.pyplot as plt
import logging

def test_subprocess_optimizer_impl(mpl_monitor):
    working_folder = mpl_monitor.runner.working_folder.replace("/", "\\") 
    test_folder = os.path.abspath(working_folder[:-3] + "test_" + working_folder[-3:])      # last 3 chars should be numeric like 000
    if os.path.exists(test_folder):
        shutil.rmtree(test_folder)
    shutil.copytree(mpl_monitor.runner.working_folder, test_folder)
    log_file = os.path.join(test_folder, "optimizer.log")
    if os.path.exists(log_file):
        os.remove(log_file)
    mpl_monitor.run_impl(mpl_monitor.optimizer, mpl_monitor.init_params, work_folder=test_folder, optimizer_test=True, debug=True, devel=True)

def test_optimizer_compatibility(optimizer, params=None):
    """ Test the compatibility of the given optimizer instance.

    This function runs a series of checks to ensure that the optimizer's methods
    and attributes are functioning as expected. It is intended for debugging and
    validation purposes.

    Parameters
    ----------
    optimizer : Optimizer
        An instance of the Optimizer class to be tested.

    init_params : array-like, optional
        Initial parameters to be used in the test. If None, the optimizer's

    Returns
    -------
    None
    """
    logger = logging.getLogger("OptimizerCompatibilityTest")
    logger.info("Starting compatibility test for optimizer: %s", optimizer.__class__.__name__)
    optimizer.prepare_for_optimization(params)
    optimizer.objective_func(params, plot=True)
    plt.show()
    logger.info("Compatibility test completed.")


"""
Optimizer.MplMonitor.py

migration of FullOptDialog to Jupyter Notebook
"""
import sys
import io
import warnings
import os
import logging
import shutil
import time
import threading
import json
from datetime import datetime
import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
from IPython.display import display, clear_output
from molass_legacy.KekLib.IpyLabelUtils import inject_label_color_css
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.KekLib.IpyLabelUtils import inject_label_color_css, set_label_color

# Global registry of active monitor instances
_ACTIVE_MONITORS = {}

class MplMonitor:
    """Interactive Jupyter notebook monitor for optimization processes with subprocess management.
    
    This class provides a dashboard-based interface for running and monitoring optimization jobs
    in Jupyter notebooks. It manages background subprocess execution, provides real-time progress
    visualization, and implements robust recovery mechanisms to prevent losing control of running
    processes when notebook outputs are cleared.
    
    The monitor tracks active processes through both an in-memory registry and a persistent file-based
    registry, allowing recovery from accidental notebook state loss.
    
    Parameters
    ----------
    function_code : str, optional
        Function code identifier for logging purposes.
    clear_jobs : bool, default=True
        If True, clears existing job folders in the optimizer directory on initialization.
    debug : bool, default=True
        If True, enables debug mode with module reloading for development.
    
    Attributes
    ----------
    optimizer_folder : str
        Path to the folder containing optimization outputs and logs.
    logger : logging.Logger
        Logger instance for recording monitor activities.
    runner : BackRunner
        Background process runner managing the subprocess execution.
    dashboard : ipywidgets.VBox
        The main dashboard widget containing plots and controls.
    process_id : str
        String representation of the current subprocess PID.
    instance_id : int
        Unique identifier for this monitor instance.
    
    Examples
    --------
    Basic usage with automatic recovery::
    
        from molass_legacy.Optimizer.MplMonitor import MplMonitor
        
        # Create and configure monitor
        monitor = MplMonitor(clear_jobs=True)
        monitor.create_dashboard()
        
        # Run optimization
        monitor.run(optimizer, init_params, niter=20, max_trials=30)
        monitor.show()
        monitor.start_watching()
    
    Recovering a lost dashboard after clearing notebook outputs::
    
        # Retrieve the most recent active monitor
        monitor = MplMonitor.get_active_monitor()
        monitor.redisplay_dashboard()
    
    Checking all active monitors::
    
        # Display status of all running monitors
        MplMonitor.show_active_monitors()
        
        # Get all active instances
        monitors = MplMonitor.get_all_active_monitors()
    
    Cleaning up orphaned processes::
    
        # Interactive cleanup of orphaned processes
        MplMonitor.cleanup_orphaned_processes()
    
    Notes
    -----
    - The monitor maintains two registries: an in-memory registry for quick access to active
      instances, and a file-based registry (``active_processes.json``) for subprocess tracking
      that persists across notebook sessions.
    
    - When creating a new monitor while others are active, a warning is displayed with
      instructions for recovery.
    
    - The dashboard includes real-time plot updates, status indicators, and control buttons
      for terminating jobs and exporting data.
    
    - Background processes are automatically cleaned up when the monitor detects they are
      orphaned or when the monitor instance is destroyed.
    
    - For optimal use in Jupyter notebooks, use ``start_watching()`` to run progress monitoring
      in a background thread, keeping the notebook interactive.
    
    .. note::
       Process registry and dashboard recovery features implemented with assistance from
       GitHub Copilot (January 2026).
    
    See Also
    --------
    BackRunner : Manages subprocess execution for optimization jobs.
    JobState : Tracks and parses optimization job state from callback files.
    
    """
    def __init__(self, function_code=None, clear_jobs=True, xr_only=False, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Optimizer.BackRunner
            reload(molass_legacy.Optimizer.BackRunner)
        from molass_legacy.Optimizer.BackRunner import BackRunner
        analysis_folder = get_setting("analysis_folder")
        optimizer_folder = os.path.join(analysis_folder, "optimized")
        self.optimizer_folder = optimizer_folder
        if clear_jobs:
            self.clear_jobs()
        logpath = os.path.join(optimizer_folder, 'monitor.log')
        self.fileh = logging.FileHandler(logpath, 'w')
        format_csv_ = '%(asctime)s,%(levelname)s,%(name)s,%(message)s'
        datefmt_ = '%Y-%m-%d %H:%M:%S'
        self.formatter_csv_ = logging.Formatter(format_csv_, datefmt_)
        self.fileh.setFormatter(self.formatter_csv_)
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(self.fileh)
        self.logger.info("MplMonitor initialized.")
        self.runner = BackRunner(xr_only=xr_only, shared_memory=False)
        self.logger.info(f"Optimizer job folder: {self.runner.optjob_folder}")
        self.result_list = []
        self.suptitle = None
        self.func_code = function_code
        self.process_id = None  # Will be set when process starts
        self.instance_id = id(self)
        self.watch_thread = None  # Will be set when watching starts
        self.stop_watch_event = threading.Event()  # For graceful thread shutdown
        
        # Check for existing active monitors and warn user
        if _ACTIVE_MONITORS:
            self.logger.info(f"Found {len(_ACTIVE_MONITORS)} existing active monitor(s)")
            print(f"⚠ Warning: {len(_ACTIVE_MONITORS)} monitor(s) already active.")
            print("  Use MplMonitor.show_active_monitors() to see them.")
            print("  Use MplMonitor.get_active_monitor() to retrieve the last one.")
        
        # Register this instance in global registry
        _ACTIVE_MONITORS[self.instance_id] = self
        self.logger.info(f"Registered monitor instance {self.instance_id}")
        
        # Clean up any orphaned processes from previous sessions
        self._cleanup_orphaned_processes()

    def clear_jobs(self):
        folder = self.optimizer_folder
        for sub in os.listdir(folder):
            subpath =  os.path.join(folder, sub)
            if os.path.isdir(subpath):
                shutil.rmtree(subpath)
                os.makedirs(subpath, exist_ok=True)

    def create_dashboard(self):
        self.plot_output = widgets.Output()

        self.status_label = widgets.Label(value="Status: Running")
        self.space_label1 = widgets.Label(value="　　　　")
        self.skip_button = widgets.Button(description="Skip Job", button_style='warning', disabled=True)
        self.space_label2 = widgets.Label(value="　　　　")
        if not hasattr(self, 'terminate_event'):
            self.terminate_event = threading.Event()
        self.terminate_button = widgets.Button(description="Terminate Job", button_style='danger')
        self.terminate_button.on_click(self.trigger_terminate)
        self.space_label3 = widgets.Label(value="　　　　")
        self.export_button = widgets.Button(description="Export Data", button_style='success', disabled=True)
        self.export_button.on_click(self.export_data)
        self.controls = widgets.HBox([self.status_label,
                                      self.space_label1,
                                      self.skip_button,
                                      self.space_label2,
                                      self.terminate_button,
                                      self.space_label3,
                                      self.export_button])

        self.message_output = widgets.Output(layout=widgets.Layout(border='1px solid gray', background_color='gray', padding='10px'))

        self.dashboard = widgets.VBox([self.plot_output, self.controls, self.message_output])
        self.dashboard_output = widgets.Output()
        self.dialog_output = widgets.Output()

    def run(self, optimizer, init_params, niter=20, seed=1234, max_trials=30, work_folder=None, dummy=False, x_shifts=None, debug=False, devel=True):
        self.optimizer = optimizer
        self.init_params = init_params
        self.nitrer = niter
        self.seed = seed
        self.num_trials = 0
        self.max_trials = max_trials
        self.work_folder = work_folder
        self.x_shifts = x_shifts
        self.run_impl(optimizer, init_params, niter=niter, seed=seed, work_folder=work_folder, dummy=dummy, debug=debug, devel=devel)

    def run_impl(self, optimizer, init_params, niter=20, seed=1234, work_folder=None, dummy=False,
                 optimizer_test=False, debug=False, devel=False):
        from importlib import reload
        import molass_legacy.Optimizer.JobState
        reload(molass_legacy.Optimizer.JobState)
        from molass_legacy.Optimizer.JobState import JobState

        if optimizer_test:
            pass
        else:
            optimizer.prepare_for_optimization(init_params)

        self.runner.run(optimizer, init_params, niter=niter, seed=seed, work_folder=work_folder, dummy=dummy, x_shifts=self.x_shifts,
                        optimizer_test=optimizer_test, debug=debug, devel=devel)
        if optimizer_test:
            abs_working_folder = os.path.abspath(work_folder)
        else:
            abs_working_folder = os.path.abspath(self.runner.working_folder)
            cb_file = os.path.join(abs_working_folder, 'callback.txt')
            self.job_state = JobState(cb_file, niter)
            # Register this process in the registry
            self._add_to_registry(abs_working_folder)
            self.curr_index = None
        self.logger.info("Starting optimization job in folder: %s with optimizer_test=%s", abs_working_folder, optimizer_test)
        
    def test_subprocess_optimizer(self):
        from importlib import reload
        import molass_legacy.Optimizer.Compatibility
        reload(molass_legacy.Optimizer.Compatibility)
        from molass_legacy.Optimizer.Compatibility import test_subprocess_optimizer_impl
        test_subprocess_optimizer_impl(self)

    def trigger_terminate(self, b):
        from molass_legacy.KekLib.IpyUtils import ask_user

        def handle_response(answer):
            print("Callback received:", answer)
            if answer:
                self.terminate_event.set()
                self.status_label.value = "Status: Terminating"
                set_label_color(self.status_label, "yellow")
                self.logger.info("Terminate job requested. id(self)=%d", id(self))
        display(self.dialog_output)
        ask_user("Do you really want to terminate?", callback=handle_response, output_widget=self.dialog_output)

    def show(self, debug=False):
        self.update_plot()
        # with self.dashboard_output:
        display(self.dashboard)
        inject_label_color_css()
        set_label_color(self.status_label, "green")

    def update_plot(self):
        from importlib import reload
        import molass_legacy.Optimizer.JobStatePlot
        reload(molass_legacy.Optimizer.JobStatePlot)
        from molass_legacy.Optimizer.JobStatePlot import plot_job_state

        # Get current plot info and best params
        plot_info = self.job_state.get_plot_info()
        params = self.get_best_params(plot_info=plot_info)

        # Prepare to capture warnings and prints
        buf_out = io.StringIO()
        buf_err = io.StringIO()
        with warnings.catch_warnings(record=True) as wlist:
            warnings.simplefilter("always")
            old_stdout, old_stderr = sys.stdout, sys.stderr
            sys.stdout = buf_out
            sys.stderr = buf_err
            try:
                with self.plot_output:
                    clear_output(wait=True)
                    plot_job_state(self, params, plot_info=plot_info, niter=self.nitrer)
                    display(self.fig)
                    plt.close(self.fig)
            finally:
                # Restore stdout/stderr
                sys.stdout = old_stdout
                sys.stderr = old_stderr

        # Unique warning messages and counts
        messages_counts = {}
        for w in wlist:
            msg = str(w.message)
            if msg in messages_counts:
                messages_counts[msg] += 1
            else:
                messages_counts[msg] = 1

        # Collect all messages
        messages = []
        # Warnings
        for msg, count in messages_counts.items():
            if count > 1:
                messages.append(f"Warning: {msg} (x {count})")
            else:
                messages.append(f"Warning: {msg}")
        # Print output and errors
        out_str = buf_out.getvalue()
        err_str = buf_err.getvalue()
        if out_str.strip():
            messages.append(out_str.strip())
        if err_str.strip():
            messages.append(err_str.strip())

        # Display all messages in message_output
        with self.message_output:
            clear_output(wait=True)
            for msg in messages:
                print(msg)

    def watch_progress(self, interval=1.0):
        """Main watching loop that monitors subprocess and updates dashboard.
        
        This runs in a background thread and can be stopped gracefully via stop_watch_event.
        """
        self.logger.info(f"Watch thread started for monitor {self.instance_id}")
        try:
            while True:
                # Check for graceful shutdown request
                if self.stop_watch_event.is_set():
                    self.logger.info("Watch thread shutdown requested")
                    break
                
                exit_loop = False
                has_ended = False
                ret = self.runner.poll()

                if ret is not None:
                    exit_loop = True
                    has_ended = True
                # self.logger.info("self.terminate=%s, id(self)=%d", str(self.terminate_event.is_set()), id(self))
                if self.terminate_event.is_set():
                    self.logger.info("Terminating optimization job.")
                    self.runner.terminate()
                    exit_loop = True

                resume_loop = False
                if exit_loop:
                    if has_ended:
                        self.logger.info("Optimization job ended normally.")
                        self.status_label.value = "Status: Completed"
                        set_label_color(self.status_label, "blue")
                        if self.num_trials < self.max_trials:
                            self.logger.info("Starting a new optimization trial (%d/%d).", self.num_trials, self.max_trials)
                            best_params = self.get_best_params()
                            self.run_impl(self.optimizer, best_params, niter=self.nitrer, seed=self.seed, work_folder=None, dummy=False, debug=False)
                            self.status_label.value = "Status: Running"
                            set_label_color(self.status_label, "green")
                            resume_loop = True
                        else:
                            self.status_label.value = "Status: Max Trials Reached"
                            set_label_color(self.status_label, "gray")
                            self.terminate_button.disabled = True
                    else:
                        self.logger.info("Optimization job terminated by user.")
                        self.status_label.value = "Status: Terminated"
                        set_label_color(self.status_label, "gray")
                        self.terminate_button.disabled = True

                    self.save_the_result_figure()
                    self.num_trials += 1

                    with self.plot_output:
                        clear_output(wait=True)  # Remove any possibly remaining plot
                    if not resume_loop:
                        # Remove from registry when fully done
                        self._remove_from_registry()
                        break

                self.job_state.update()
                if self.job_state.has_changed():
                    self.update_plot()
                    # clear_output(wait=True)
                    # display(self.dashboard)
                time.sleep(interval)
        finally:
            self.watch_thread = None
            self.logger.info(f"Watch thread ended for monitor {self.instance_id}")

    def start_watching(self):
        """Start the background thread that monitors optimization progress.
        
        Only one watch thread can be active per monitor instance. If a thread is
        already running, this method will log a warning and return without starting
        a new thread.
        """
        # Check if thread is already running
        if self.watch_thread is not None and self.watch_thread.is_alive():
            self.logger.warning(f"Watch thread already running for monitor {self.instance_id}")
            print("⚠ Warning: Watch thread is already running for this monitor.")
            return
        
        # Clear stop event in case it was set previously
        self.stop_watch_event.clear()
        
        # Avoid Blocking the Main Thread:
        # Never run a long or infinite loop in the main thread in Jupyter if you want widget interactivity.
        self.watch_thread = threading.Thread(target=self.watch_progress, daemon=True)
        self.watch_thread.start()
        self.logger.info(f"Started watch thread for monitor {self.instance_id}")
    
    def stop_watching(self, timeout=5.0):
        """Stop the background watch thread gracefully.
        
        Args:
            timeout: Maximum time in seconds to wait for thread to stop.
        
        Returns:
            bool: True if thread stopped successfully, False if timeout occurred.
        """
        if self.watch_thread is None or not self.watch_thread.is_alive():
            self.logger.info("No active watch thread to stop")
            return True
        
        self.logger.info(f"Stopping watch thread for monitor {self.instance_id}")
        self.stop_watch_event.set()
        self.watch_thread.join(timeout=timeout)
        
        if self.watch_thread.is_alive():
            self.logger.warning(f"Watch thread did not stop within {timeout}s")
            return False
        else:
            self.logger.info("Watch thread stopped successfully")
            self.watch_thread = None
            return True
    
    def is_watching(self):
        """Check if the watch thread is currently active.
        
        Returns:
            bool: True if watch thread is running, False otherwise.
        """
        return self.watch_thread is not None and self.watch_thread.is_alive()
    
    def get_best_params(self, plot_info=None):
        if plot_info is None:
            plot_info = self.job_state.get_plot_info()

        x_array = plot_info[-1]

        if len(x_array) == 0:
            self.curr_index = 0
            return self.init_params

        fv = plot_info[0]
        k = np.argmin(fv[:,1])
        self.curr_index = k
        best_params = x_array[k]
        return best_params

    def save_the_result_figure(self, fig_file=None):
        if fig_file is None:
            figs_folder = os.path.join(self.optimizer_folder, "figs")
            if not os.path.exists(figs_folder):
                os.makedirs(figs_folder)
            fig_file = os.path.join(figs_folder, "fig-%03d.jpg" % self.num_trials)
        self.fig.savefig(fig_file)

    def export_data(self, b, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.LrfExporter
            reload(Optimizer.LrfExporter)
        from .LrfExporter import LrfExporter

        params = self.optimizer.init_params
        try:
            exporter = LrfExporter(self.optimizer, params, self.dsets)
            folder = exporter.export()
            fig_file = os.path.join(folder, "result_fig.jpg")
            self.save_the_result_figure(fig_file=fig_file)
            print(f"Exported to folder: {folder}")
        except Exception as exc:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "export: ")
            print(f"Failed to export due to: {exc}")

    # ===== Dashboard Recovery =====
    
    def redisplay_dashboard(self):
        """Redisplay the dashboard after it has been cleared.
        
        This method allows you to reconnect to a running monitor after
        accidentally clearing notebook outputs. Call this to get the
        dashboard back.
        
        Example:
            # After clearing outputs, retrieve and redisplay:
            monitor = MplMonitor.get_active_monitor()
            monitor.redisplay_dashboard()
        """
        if not hasattr(self, 'dashboard'):
            print("Dashboard not initialized. Call create_dashboard() first.")
            return
        
        # Update plot with current state
        if hasattr(self, 'job_state'):
            self.update_plot()
        
        # Redisplay the dashboard
        display(self.dashboard)
        inject_label_color_css()
        
        # Restore status label color based on current status
        status = self.status_label.value
        if "Running" in status:
            set_label_color(self.status_label, "green")
        elif "Completed" in status:
            set_label_color(self.status_label, "blue")
        elif "Terminated" in status or "Max Trials" in status:
            set_label_color(self.status_label, "gray")
        elif "Terminating" in status:
            set_label_color(self.status_label, "yellow")
        
        print(f"Dashboard redisplayed for monitor {self.instance_id}")
        self.logger.info(f"Dashboard redisplayed for instance {self.instance_id}")
    
    @classmethod
    def get_active_monitor(cls):
        """Get the most recently created active monitor instance.
        
        Returns the last MplMonitor instance that was created and is still active.
        Useful for recovering access to a monitor after clearing notebook outputs.
        
        Returns:
            MplMonitor: The most recent active monitor, or None if no monitors exist.
        
        Example:
            # After clearing outputs:
            monitor = MplMonitor.get_active_monitor()
            if monitor:
                monitor.redisplay_dashboard()
        """
        if not _ACTIVE_MONITORS:
            print("No active monitors found.")
            return None
        
        # Return the most recent (last inserted) monitor
        return list(_ACTIVE_MONITORS.values())[-1]
    
    @classmethod
    def get_all_active_monitors(cls):
        """Get all active monitor instances.
        
        Returns:
            list: List of all active MplMonitor instances.
        """
        return list(_ACTIVE_MONITORS.values())
    
    @classmethod
    def show_active_monitors(cls):
        """Display information about all active monitor instances.
        
        Shows a summary of all currently active monitors including their
        status, process ID, and working folder if available.
        
        Example:
            MplMonitor.show_active_monitors()
        """
        if not _ACTIVE_MONITORS:
            print("No active monitors found.")
            return
        
        print(f"Found {len(_ACTIVE_MONITORS)} active monitor(s):\n")
        
        for idx, (instance_id, monitor) in enumerate(_ACTIVE_MONITORS.items(), 1):
            print(f"Monitor #{idx} (ID: {instance_id})")
            
            # Status
            if hasattr(monitor, 'status_label'):
                print(f"  Status: {monitor.status_label.value}")
            else:
                print(f"  Status: Not started")
            
            # Process info
            if hasattr(monitor, 'process_id') and monitor.process_id:
                print(f"  Process ID: {monitor.process_id}")
            
            # Thread info
            if hasattr(monitor, 'watch_thread'):
                if monitor.watch_thread is not None and monitor.watch_thread.is_alive():
                    print(f"  Watch Thread: ACTIVE (ID: {monitor.watch_thread.ident})")
                else:
                    print(f"  Watch Thread: NOT RUNNING")
            
            # Working folder
            if hasattr(monitor, 'runner') and hasattr(monitor.runner, 'working_folder'):
                print(f"  Working folder: {monitor.runner.working_folder}")
            
            # Trial info
            if hasattr(monitor, 'num_trials') and hasattr(monitor, 'max_trials'):
                print(f"  Trials: {monitor.num_trials}/{monitor.max_trials}")
            
            print()
        
        print("To redisplay a dashboard:")
        print("  monitor = MplMonitor.get_active_monitor()")
        print("  monitor.redisplay_dashboard()")
    
    @classmethod
    def cleanup_orphaned_threads(cls):
        """Stop watch threads for monitors that are no longer needed.
        
        This method identifies and stops watch threads that are still running
        for monitors that may have lost their dashboard. Useful for cleaning up
        after accidentally clearing notebook outputs multiple times.
        
        Example:
            MplMonitor.cleanup_orphaned_threads()
        """
        if not _ACTIVE_MONITORS:
            print("No active monitors found.")
            return
        
        orphaned_count = 0
        stopped_count = 0
        
        for instance_id, monitor in _ACTIVE_MONITORS.items():
            if hasattr(monitor, 'watch_thread') and monitor.watch_thread is not None:
                if monitor.watch_thread.is_alive():
                    orphaned_count += 1
                    print(f"Monitor {instance_id}: Watch thread is running")
                    
                    # Check if we should stop it
                    response = input("  Stop this watch thread? (y/n): ").strip().lower()
                    if response == 'y':
                        print("  Stopping thread...")
                        success = monitor.stop_watching(timeout=5.0)
                        if success:
                            print("  Thread stopped successfully.")
                            stopped_count += 1
                        else:
                            print("  Warning: Thread did not stop cleanly.")
        
        if orphaned_count == 0:
            print("No active watch threads found.")
        else:
            print(f"\nStopped {stopped_count} of {orphaned_count} watch thread(s).")

    # ===== Process Registry Management =====
    
    def _get_registry_path(self):
        """Get the path to the process registry file."""
        return os.path.join(self.optimizer_folder, 'active_processes.json')
    
    def _load_registry(self):
        """Load the process registry from disk."""
        registry_path = self._get_registry_path()
        if not os.path.exists(registry_path):
            return {}
        try:
            with open(registry_path, 'r') as f:
                return json.load(f)
        except (json.JSONDecodeError, IOError) as e:
            self.logger.warning(f"Failed to load registry: {e}")
            return {}
    
    def _save_registry(self, registry):
        """Save the process registry to disk."""
        registry_path = self._get_registry_path()
        try:
            with open(registry_path, 'w') as f:
                json.dump(registry, f, indent=2)
        except IOError as e:
            self.logger.error(f"Failed to save registry: {e}")
    
    def _add_to_registry(self, working_folder):
        """Add current process to the registry."""
        if not hasattr(self.runner, 'process') or self.runner.process is None:
            self.logger.warning("No process to register")
            return
        
        pid = self.runner.process.pid
        self.process_id = str(pid)
        registry = self._load_registry()
        registry[self.process_id] = {
            'pid': pid,
            'working_folder': working_folder,
            'timestamp': datetime.now().isoformat(),
            'status': 'running'
        }
        self._save_registry(registry)
        self.logger.info(f"Registered process PID {pid} in registry")
    
    def _remove_from_registry(self):
        """Remove current process from the registry."""
        if self.process_id is None:
            return
        
        registry = self._load_registry()
        if self.process_id in registry:
            del registry[self.process_id]
            self._save_registry(registry)
            self.logger.info(f"Removed process {self.process_id} from registry")
        self.process_id = None
    
    def _is_process_alive(self, pid):
        """Check if a process with given PID is alive."""
        try:
            import psutil
            return psutil.pid_exists(pid)
        except ImportError:
            # Fallback to OS-specific method if psutil not available
            import signal
            if os.name == 'nt':  # Windows
                import subprocess
                try:
                    result = subprocess.run(
                        ['tasklist', '/FI', f'PID eq {pid}'],
                        capture_output=True, text=True, timeout=2
                    )
                    return str(pid) in result.stdout
                except Exception:
                    return False
            else:  # Unix-like
                try:
                    os.kill(pid, 0)
                    return True
                except OSError:
                    return False
    
    def _cleanup_orphaned_processes(self, auto_terminate=True):
        """Clean up orphaned processes from previous sessions.
        
        Args:
            auto_terminate: If True, automatically terminate orphaned processes.
                           If False, only report them.
        """
        registry = self._load_registry()
        if not registry:
            return
        
        orphaned = []
        cleaned = []
        
        for proc_id, info in list(registry.items()):
            pid = info.get('pid')
            if pid is None:
                cleaned.append(proc_id)
                continue
            
            # Check if process is still alive
            if not self._is_process_alive(pid):
                self.logger.info(f"Process {pid} is no longer running, removing from registry")
                cleaned.append(proc_id)
            else:
                orphaned.append(info)
                if auto_terminate:
                    self.logger.warning(f"Terminating orphaned process {pid} from {info.get('timestamp')}")
                    try:
                        import psutil
                        proc = psutil.Process(pid)
                        proc.terminate()
                        try:
                            proc.wait(timeout=5)
                        except psutil.TimeoutExpired:
                            proc.kill()
                        cleaned.append(proc_id)
                    except ImportError:
                        # Fallback without psutil
                        if os.name == 'nt':  # Windows
                            import subprocess
                            subprocess.run(['taskkill', '/F', '/PID', str(pid)], 
                                         capture_output=True)
                        else:  # Unix-like
                            import signal
                            os.kill(pid, signal.SIGTERM)
                        cleaned.append(proc_id)
                    except Exception as e:
                        self.logger.error(f"Failed to terminate process {pid}: {e}")
        
        # Clean up registry
        for proc_id in cleaned:
            if proc_id in registry:
                del registry[proc_id]
        
        if cleaned:
            self._save_registry(registry)
            self.logger.info(f"Cleaned up {len(cleaned)} entries from process registry")
        
        if orphaned and not auto_terminate:
            print(f"Warning: Found {len(orphaned)} orphaned processes:")
            for info in orphaned:
                print(f"  - PID {info['pid']} started at {info['timestamp']}")
            print("Call MplMonitor.cleanup_orphaned_processes() to terminate them.")
    
    @classmethod
    def cleanup_orphaned_processes(cls, optimizer_folder=None):
        """Class method to manually clean up orphaned processes.
        
        This can be called from a fresh notebook cell without an instance:
            MplMonitor.cleanup_orphaned_processes()
        
        Args:
            optimizer_folder: Path to optimizer folder. If None, uses default from settings.
        """
        if optimizer_folder is None:
            analysis_folder = get_setting("analysis_folder")
            optimizer_folder = os.path.join(analysis_folder, "optimized")
        
        registry_path = os.path.join(optimizer_folder, 'active_processes.json')
        
        if not os.path.exists(registry_path):
            print("No active process registry found.")
            return
        
        try:
            with open(registry_path, 'r') as f:
                registry = json.load(f)
        except (json.JSONDecodeError, IOError) as e:
            print(f"Failed to load registry: {e}")
            return
        
        if not registry:
            print("No active processes in registry.")
            return
        
        print(f"Found {len(registry)} process(es) in registry:")
        
        for proc_id, info in list(registry.items()):
            pid = info.get('pid')
            timestamp = info.get('timestamp', 'unknown')
            working_folder = info.get('working_folder', 'unknown')
            
            # Check if process is alive
            try:
                import psutil
                is_alive = psutil.pid_exists(pid)
            except ImportError:
                # Fallback
                if os.name == 'nt':
                    import subprocess
                    try:
                        result = subprocess.run(
                            ['tasklist', '/FI', f'PID eq {pid}'],
                            capture_output=True, text=True, timeout=2
                        )
                        is_alive = str(pid) in result.stdout
                    except Exception:
                        is_alive = False
                else:
                    try:
                        os.kill(pid, 0)
                        is_alive = True
                    except OSError:
                        is_alive = False
            
            if is_alive:
                print(f"  - PID {pid}: RUNNING (started {timestamp})")
                print(f"    Folder: {working_folder}")
                response = input(f"    Terminate this process? (y/n): ").strip().lower()
                if response == 'y':
                    try:
                        import psutil
                        proc = psutil.Process(pid)
                        proc.terminate()
                        try:
                            proc.wait(timeout=5)
                            print(f"    Terminated PID {pid}")
                        except psutil.TimeoutExpired:
                            proc.kill()
                            print(f"    Killed PID {pid} (did not respond to terminate)")
                        del registry[proc_id]
                    except ImportError:
                        if os.name == 'nt':
                            import subprocess
                            subprocess.run(['taskkill', '/F', '/PID', str(pid)])
                        else:
                            import signal
                            os.kill(pid, signal.SIGTERM)
                        print(f"    Terminated PID {pid}")
                        del registry[proc_id]
                    except Exception as e:
                        print(f"    Failed to terminate: {e}")
            else:
                print(f"  - PID {pid}: NOT RUNNING (removing from registry)")
                del registry[proc_id]
        
        # Save updated registry
        try:
            with open(registry_path, 'w') as f:
                json.dump(registry, f, indent=2)
            print("\nRegistry updated.")
        except IOError as e:
            print(f"Failed to save registry: {e}")
    
    def __del__(self):
        """Cleanup when object is destroyed."""
        # Stop watch thread gracefully
        try:
            if hasattr(self, 'watch_thread') and self.watch_thread is not None:
                if self.watch_thread.is_alive():
                    if hasattr(self, 'logger'):
                        self.logger.info(f"Stopping watch thread in __del__ for {self.instance_id}")
                    self.stop_watch_event.set()
                    self.watch_thread.join(timeout=2.0)
        except Exception:
            pass  # Ignore errors during cleanup
        
        try:
            self._remove_from_registry()
        except Exception:
            pass  # Ignore errors during cleanup
        
        # Remove from global instance registry
        try:
            if hasattr(self, 'instance_id') and self.instance_id in _ACTIVE_MONITORS:
                del _ACTIVE_MONITORS[self.instance_id]
                if hasattr(self, 'logger'):
                    self.logger.info(f"Unregistered monitor instance {self.instance_id}")
        except Exception:
            pass  # Ignore errors during cleanup





"""
Optimizer.Scripting.py

- Menus/V2Menu.py
- Optimizer/OptimizerUtils.py
- Optimizer/OptStrategyDialog.py
- Peaks/PeakEditor.py
- Optimizer/FullOptDialog.py

- Optimizer/optimizer.py
    - Optimizer/OptimizerMain.py
"""
import os
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import set_setting

def prepare_data(in_folder, sd=None, clear_temp_settings=True, analysis_folder=None):
    from molass_legacy.Global.V2Init import update_sec_settings

    if clear_temp_settings:
        from molass_legacy._MOLASS.SerialSettings import set_setting, clear_v2_temporary_settings
        clear_v2_temporary_settings()

    set_setting('in_folder', in_folder)
    if analysis_folder is None:
        analysis_folder = 'temp_analysis'
        if not os.path.exists(analysis_folder):
            os.makedirs(analysis_folder)
    set_setting('analysis_folder', analysis_folder)
    update_sec_settings()

    optimizer_folder = os.path.join(analysis_folder, "optimized")
    if not os.path.exists(optimizer_folder):
        os.makedirs(optimizer_folder)
    rg_folder = os.path.join(optimizer_folder, "rg-curve")
    if not os.path.exists(rg_folder):
        os.makedirs(rg_folder)

    if sd is None:
        from molass_legacy.Batch.StandardProcedure import StandardProcedure
        sp = StandardProcedure()
        sd = sp.load_old_way(in_folder)

    if sd.pre_recog is None:
        from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
        sd.pre_recog = PreliminaryRecognition(sd)

    from molass_legacy.Batch.OptDataSetsProxy import OptDataSetsProxy as OptDataSets
    from molass_legacy.SecSaxs.DataTreatment import DataTreatment

    trimming = 2
    correction = 1
    unified_baseline_type = 1
    treat = DataTreatment(route="v2", trimming=trimming, correction=correction, unified_baseline_type=unified_baseline_type)
    pre_recog = sd.pre_recog
    trimmed_sd = treat.get_trimmed_sd(sd, pre_recog)
    corrected_sd = treat.get_corrected_sd(sd, pre_recog, trimmed_sd)
    treat.save()

    from molass_legacy.Batch.FullBatch import FullBatch
    from molass_legacy.Optimizer.FullOptInput import FullOptInput
    # equivalent to PeakEditor.__init__
    batch = FullBatch()
    batch.logger = logging.getLogger(__name__)
    batch.sd = trimmed_sd
    batch.corrected_sd = corrected_sd
    batch.pre_recog = PreliminaryRecognition(corrected_sd)  # to ensure better that mapping is available
    batch.base_curve_info = treat.get_base_curve_info()     # not used?

    batch.strict_sec_penalty = False
    batch.fullopt_class, batch.class_code = None, None
    batch.fullopt_input = FullOptInput(sd=trimmed_sd, corrected_sd=corrected_sd, rg_folder=rg_folder)
    batch.dsets = batch.fullopt_input.get_dsets(progress_cb=None, compute_rg=True, possibly_relocated=False)
    return batch

def prepare_optimizer(batch, num_components=3, model="EGH", method="BH", function_code=None, debug=False):
    from molass_legacy.Optimizer.FuncImporter import import_objective_function

    if function_code is None:
        assert model is not None, "Either model or function_code must be provided."
        from molass_legacy.Optimizer.OptimizerUtils import get_function_code
        function_code = get_function_code(model)
    else:
        assert model is None, "Either model or function_code must be provided, not both."
        from molass_legacy.Optimizer.OptimizerUtils import get_model_name       
        model_name = get_model_name(function_code)

    function_class = import_objective_function(function_code)
    set_optimizer_settings(num_components=num_components, model=model, method=method)

    if debug:
        print("Running optimizer with function:", function_class.__name__)

    batch.exact_num_peaks = num_components

    # equivalent to PeakEditor.body
    uv_x, uv_y, xr_x, xr_y, baselines = batch.get_curve_xy(return_baselines=True)
    uv_y_ = uv_y - baselines[0]
    xr_y_ = xr_y - baselines[1]
    uv_peaks, xr_peaks = batch.get_modeled_peaks(uv_x, uv_y_, xr_x, xr_y_)
    batch.set_lrf_src_args1(uv_x, uv_y, xr_x, xr_y, baselines)

    batch.construct_optimizer(fullopt_class=function_class)
    return batch.optimizer

def set_optimizer_settings(num_components=3, model="EGH", method="BH", param_init_type=1):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from .OptimizerSettings import OptimizerSettings

    elution_model = 0
    model = model.upper()
    if model == "EGH":
        elution_model = 0
    elif model == "SDM":
        elution_model = 2
    elif model == "EDM":
        elution_model = 5
    else:
        assert False, f"Unknown model: {model}"

    solver_name = method.upper()
    if solver_name == "BH":
        optimization_method = 0
    elif solver_name == "NS":
        optimization_method = 1
    else:
        assert False, f"Unknown method: {method}"
    set_setting("optimization_method", optimization_method)     # for backward compatibility

    separate_eoii_flags = [0] * num_components

    settings = OptimizerSettings(param_init_type=param_init_type,
                                 elution_model=elution_model,
                                 optimization_method=optimization_method,
                                 separate_eoii_flags=separate_eoii_flags)
    settings.save()

def estimate_init_params(batch, optimizer, developing=False, debug=False):
    batch.get_ready_for_progress_display()

    init_params = batch.compute_init_params(developing=developing, debug=debug)
    print("Initial Parameters:", len(init_params))
    optimizer.prepare_for_optimization(init_params)
    return init_params

def run_optimizer(optimizer, init_params, niter=20, clear_jobs=True, dummy=False, x_shifts=None, debug=True):
    """ Run the optimizer with a monitoring dashboard.

    Parameters
    ----------
    optimizer : Optimizer
        The optimizer instance to run.
    init_params : array-like
        Initial parameters for the optimization.
    niter : int, optional
        Number of iterations to run. Default is 20.
    clear_jobs : bool, optional
        Whether to clear previous jobs. Default is True.
    dummy : bool, optional
        If True, runs in dummy mode without actual optimization. Default is False.
    x_shifts : array-like, optional
        Shifts to apply to the parameters. Default is None.
        x_shifts can be obtained from dsets.get_x_shifts().
        See also OptDataSets.get_x_shifts().
    debug : bool, optional
        If True, enables debug mode. Default is True.

    Returns
    -------
    monitor : MplMonitor
        The monitoring dashboard instance.
    """
    if debug:
        from importlib import reload
        import molass_legacy.Optimizer.MplMonitor
        reload(molass_legacy.Optimizer.MplMonitor)
    from molass_legacy.Optimizer.MplMonitor import MplMonitor
    monitor = MplMonitor(optimizer.get_function_code(), clear_jobs=clear_jobs, xr_only=optimizer.get_xr_only())
    monitor.create_dashboard()
    monitor.run(optimizer, init_params, niter=niter, dummy=dummy, x_shifts=x_shifts, debug=debug)
    monitor.show(debug=debug)

    if dummy:
        import time
        time.sleep(10)
        monitor.terminate_job(None)
    else:
        monitor.start_watching()
    return monitor

def get_params(job_result_folder, index=None, debug=False):
    from .StateSequence import read_callback_txt_impl
    cb_file = os.path.join(job_result_folder, 'callback.txt')
    fv_list, x_list = read_callback_txt_impl(cb_file)
    fv = np.array(fv_list)
    x = np.array(x_list)
    if index is None:
        k = np.argmin(fv[:,1])
        params = x[k]
        if debug:
            print("Best parameters at index %d with fv=%g" % (k, fv[k,1]))
    else:
        params = x[index]
        if debug:
            print("Parameters at index %d with fv=%g" % (index, fv[index,1]))
    return params


"""
KekLib.IpyLabelUtils.py
"""

from IPython.display import display, HTML

def inject_label_color_css():
    """
    This function adds CSS styles for different label colors (green, red, yellow, gray)
    to be used with ipywidgets.Label.

    Parameters:
    -----------
    None

    Returns:
    --------
    None
    """

    display(HTML("""
    <style>
    .status-label-green {
        color: green !important;
        font-weight: bold;
    }
    .status-label-red {
        color: red !important;
        font-weight: bold;
    }
    .status-label-yellow {
        color: orange !important;
        font-weight: bold;
    }
    .status-label-yellow {
        color: orange !important;
        font-weight: bold;
    }
    .status-label-gray {
        color: gray !important;
        font-weight: bold;
    }
    .status-label-blue {
        color: blue !important;     
        font-weight: bold;
    }
    </style>
    """))

def set_label_color(label, color):
    """
    Set the color of an ipywidgets.Label by adding/removing CSS classes.

    Parameters:
    -----------
    label : ipywidgets.Label
        The label widget to modify.
    color : str
        The color to set. Must be one of: "green", "red", "yellow", "gray", "blue".
        
    Returns:
    --------
    None
    """
    if color not in ["green", "red", "yellow", "gray", "blue"]:
        raise ValueError("Color must be one of: green, red, yellow, gray, blue.")
    
    # Remove all color classes
    for c in ["status-label-green", "status-label-red", "status-label-yellow", "status-label-gray", "status-label-blue"]:
        label.remove_class(c)
    # Add the selected color class
    label.add_class(f"status-label-{color}")


"""
Progress chart for optimization GUI.
"""
from datetime import timedelta
from molass_legacy.KekLib.NumpyUtils import get_proportional_points
from molass_legacy.KekLib.TimeUtils import friendly_time_str
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from .FvScoreConverter import convert_score

SHOW_FV_MIN = 0
SHOW_FV_MAX = 100
PROGRESS_X_MARGIN = 1000    # not too large for xmax = 500000
DEVELOP_MODE = True
ENABLE_DEVEL_POPUP_WHILE_BUSY = True

def get_xlim_prog_axes(max_num_evals):
    xmin = -PROGRESS_X_MARGIN
    xmax = max_num_evals + PROGRESS_X_MARGIN
    return xmin, xmax

def get_time_started(fv_array):
    try:
        start_time = fv_array[0, 3]
        time = friendly_time_str(start_time)
    except:
        # IndexError: index 3 is out of bounds for axis 1 with size 2
        time = ""
    return time

def get_time_elapsed(fv_array):
    try:
        start_time = fv_array[0, 3]
        curr_time = fv_array[-1, 3]
        hhmmss = str(curr_time - start_time).split(":")
        time = "%3d.%02d" % tuple([int(s) for s in hhmmss[0:2]])
        # %3d instead of %2d is just for positioning purpose with non-fixed-width fonts.
    except:
        log_exception(None, "get_time_elapsed: ")
        # IndexError: index 3 is out of bounds for axis 1 with size 2
        time = ""
    return time

def guess_ending_time(fv_array):
    ending_time = ""
    return ending_time

def guess_ending_time(fv_array, niter=20):
    finish_time = None
    time = ""

    if fv_array.shape[0] > 3:
        # unreliable when fv.shape[0] <= 3
        try:
            start_time = fv_array[0,3]
            curr_time = fv_array[-1,3]
            finish_time = start_time + (curr_time - start_time)*(niter/fv_array.shape[0])
            # add 1 minute so that it won't be too early
            time = friendly_time_str(finish_time + timedelta(minutes=1))
        except:
            pass

    return time, finish_time

def get_remaining_time(fv_array, finish_time):
    if finish_time is None:
        return ""
    try:
        curr_time = fv_array[-1,3]
        # add 1 minute so that it won't be too short
        hhmmss = str(finish_time - curr_time + timedelta(minutes=1) ).split(":")
        time = "%3d.%02d" % tuple([int(s) for s in hhmmss[0:2]])
        # %3d instead of %2d is just for positioning purpose with non-fixed-width fonts.
        # to be fixed: ValueError: invalid literal for int() with base 10: '-1 day, 23'
    except:
        # log_exception(self.logger, "get_remaining_time: ")
        # IndexError: index 3 is out of bounds for axis 1 with size 2
        # ValueError: invalid literal for int() with base 10: '-1 day, 23'
        time = ""
    return time

def draw_progress(self, plot_info, niter=20):

    for ax in self.prog_axes:
        ax.cla()
    for ax in self.prog_axes[0:3]:
        ax.set_xticklabels([])

    fv, max_num_evals, x_array = plot_info
    if fv.shape[0] == 0:
        return

    # Function Values
    prog_ax = self.prog_ax

    x_, y_ = fv[:,0:2].T
    prog_ax.plot(x_, convert_score(y_))
    prog_ax.set_xlim(-PROGRESS_X_MARGIN, max_num_evals + PROGRESS_X_MARGIN)
    ymin, ymax = prog_ax.get_ylim()
    prog_ax.set_ylim(ymin, ymax)     # these limits will be reset below

    # m = self.best_index
    m = len(x_array) - 1
    if self.curr_index is None:
        self.curr_index = m
    ymin, ymax = prog_ax.get_ylim()
    prog_ax.set_ylim(max(SHOW_FV_MIN, ymin), SHOW_FV_MAX)

    params_type = self.optimizer.params_type

    # Peak Top Positions
    peak_ax = self.prog_axes[1]
    n = self.optimizer.n_components
    x_ = fv[:,0]
    pos_array_list = params_type.get_peak_pos_array_list(x_array)
    for y_ in pos_array_list:
        peak_ax.plot(x_, y_)

    xmin, xmax = self.axes[1].get_xlim()
    xmin_, xmax_ = get_proportional_points(xmin, xmax, [-0.1, 1.1])
    peak_ax.set_ylim(xmin_, xmax_)

    # Rg Values
    rg_ax = self.prog_axes[2]
    rg_start = params_type.get_rg_start_index()
    for k in range(n):
        y_ = x_array[:,rg_start+k]
        rg_ax.plot(x_, y_)
    ymin_rg, ymax_rg = rg_ax.get_ylim()
    rg_ax.set_ylim(10, ymax_rg*1.2)

    # Mapped Range
    map_ax = self.prog_axes[3]
    mr_start = params_type.get_mr_start_index()
    for k in range(2):
        y_ = x_array[:,mr_start+k]
        map_ax.plot(x_, y_)
    map_ax.set_ylim(xmin_, xmax_)

    # Best and Current Result Indicator
    xmin, xmax = get_xlim_prog_axes(max_num_evals)
    best_x = fv[m,0]
    for k, ax in enumerate(self.prog_axes):
        if k > 0:
            ax.set_xlim(xmin, xmax)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
        ax.plot([best_x, best_x], [ymin, ymax], color='red')
        if self.curr_index != m:
            x = fv[self.curr_index,0]
            ax.plot([x, x], [ymin, ymax], color='yellow')

        if fv.shape[0]-1 not in [m, self.curr_index]:
            x = fv[-1,0]
            ax.plot([x, x], [ymin, ymax], color='gray', alpha=0.3)

    ymin_, ymax_ = map_ax.get_ylim()
    dy = (ymax_ - ymin_) * 1.0
    ymin, ymax = ymin_ + dy, ymax_ + dy
    tx = xmax*1.07

    w = 2.2
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, "Starting Time", ha="center")

    time_str = get_time_started(fv)
    w = 2.0
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, time_str, ha="center", va="center")

    w = 1.6
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, "Time Elapsed", ha="center")

    time_str = get_time_elapsed(fv)
    w = 1.4
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, time_str, ha="center", va="center")

    w = 0.4
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, "Ending Time", ha="center")

    # guess_ending_time() must be called before get_remaining_time()
    time_str, finish_time = guess_ending_time(fv, niter=niter)
    w = 0.2
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, time_str, ha="center", va="center")

    w = 1.0
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, "Time Ahead", ha="center")

    time_str = get_remaining_time(fv, finish_time)
    w = 0.8
    ty = ymin*(1-w) + ymax*w
    map_ax.text(tx, ty, time_str, ha="center", va="center")


"""
Optimizer.JobState.py
"""
import os
import logging
import numpy as np
import time
from .StateSequence import read_callback_txt_impl
from .OptimizerUtils import get_method_name

class JobState:
    def __init__(self, cb_file, niter=20):
        self.logger = logging.getLogger(__name__)
        self.cb_file = cb_file
        self.solver_name = get_method_name()
        self.niter = niter
        self.last_mod_time = None
        self._has_changed = False
        self.fv = np.array([])
        self.x = np.array([])
        self.xmax = None
        self.time_created = time.time()

    def has_changed(self):
        return self._has_changed

    def get_plot_info(self):
        return self.fv, self.xmax, self.x

    def estimate_xmax(self, fv_list):
        counter = fv_list[-1][0] if len(fv_list) > 0 else 0
        if counter == 0:
            # init state
            xmax = 500000*self.niter//100
        else:
            xmax = int(counter * (self.niter+1)/len(fv_list))
        return xmax

    def update(self, debug=True):
        if not os.path.exists(self.cb_file):
            t = time.time() - self.time_created
            if t > 10:
                self.logger.warning("callback.txt file not found: %s", self.cb_file)
            return
        
        mod_time = os.path.getmtime(self.cb_file)
        if self.last_mod_time is not None and mod_time == self.last_mod_time:
            self._has_changed = False
            return  # No changes
        
        self.last_mod_time = mod_time
        self._has_changed = True

        fv_list, x_list = read_callback_txt_impl(self.cb_file)

        if debug:
                self.logger.info("updating information from %s, len(x_list)=%d", self.cb_file, len(x_list))

        self.fv = np.array(fv_list)
        self.x = np.array(x_list)

        if self.solver_name == "ultranest":
            from molass_legacy.Solvers.UltraNest.SolverUltraNest import get_max_ncalls
            # task: unify this estimation
            xmax = get_max_ncalls(self.niter)
        else:
            xmax = self.estimate_xmax(fv_list)

        self.xmax = xmax


import os
import openpyxl
from openpyxl.chart import LineChart, Reference

# 1. openpyxl で Excel ファイル作成
wb = openpyxl.Workbook()
ws = wb.active
for i in range(1, 11):
    ws.append([i, i * 2])

chart = LineChart()
data = Reference(ws, min_col=2, min_row=1, max_row=10)
chart.add_data(data, titles_from_data=False)
ws.add_chart(chart, "E5")
wb.save("test_chart.xlsx")

# 2. pywin32 で Excel を開く
import win32com.client

xlApp = win32com.client.Dispatch("Excel.Application")
xlApp.Visible = False
this_dir = os.path.dirname(os.path.abspath(__file__))
book_path = os.path.join(this_dir, "test_chart.xlsx")
xlBook = xlApp.Workbooks.Open(book_path)
xlSheet = xlBook.Worksheets(1)

# 3. グラフオブジェクト取得
chart_obj = xlSheet.ChartObjects(1)  # 1始まり
chart = chart_obj.Chart

# 4. y軸取得
xlValue = 2  # Excel 定数
y_axis = chart.Axes(xlValue)

# 5. TickLabels の存在確認
try:
    tick_labels = y_axis.TickLabels
    print("TickLabels:", tick_labels)
    # 例: 書式設定を試す
    tick_labels.NumberFormatLocal = "0.0E+00"
    print("NumberFormatLocal 設定成功")
except Exception as e:
    print("TickLabels アクセス時に例外:", e)

# 6. 後始末
xlBook.Close(SaveChanges=False)
xlApp.Quit()


"""

    Version.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF

"""
import platform
from molass_legacy import get_version

def get_version_string(cpuid=False, with_date=False):
    """
    Get MOLASS version string.

    Parameters
    ----------
    cpuid : bool
        If True, append cpuid information to the version string.
    with_date : bool
        If True, append the build date to the version string.
        This option is for backward compatibility.
        To keep it as it was before, update this file's timestamp on each release.

    Returns
    -------
    str
        The version string.
    """
    if cpuid:
        from molass_legacy.KekLib.MachineTypes import get_cpuid
        cpuid = ' cpuid:' + str(get_cpuid())
    else:
        cpuid = ''

    version = get_version()
    if with_date:
        import os
        from datetime import datetime
        timestamp_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'pyproject.toml')
        if not os.path.exists(timestamp_file):
            timestamp_file = __file__
        timestamp = os.path.getmtime(timestamp_file)
        version_date = datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S ")
    else:
        version_date = ""

    return'MOLASS %s (%spython %s %s%s)' % (
                version, version_date, platform.python_version(), platform.architecture()[0], cpuid )

def molass_version_for_publication():
    import re
    version = get_version_string()
    return re.sub(r"\s+\(.+", "", version)

def is_developing_version():
    return get_version_string().find(":") > 0
